from __future__ import annotations

import pickle

from dxtbx.format.image import ImageBool
from iotbx.phil import parse

import dials.util

help_message = """

This program augments a experiments JSON file with one or more masks specified by the
user.  Its only function is to input the mask file paths to the experiments JSON file,
but means that the user does not have to edit the experiments file by hand.

Crucially, the mask files must be provided in the same order as their corresponding
imagesets (sequences) appear in the experiments JSON file.

Examples::

    dials.apply_mask models.expt input.mask=pixels.mask

    dials.apply_mask models.expt input.mask=pixels1.mask input.mask=pixels2.mask
"""

phil_scope = parse(
    """
        input {
            mask = None
                .multiple = True
                .type = str
                .help = "The mask filenames, one mask per imageset"
        }

        output {
            experiments = masked.expt
                .type = str
                .help = "Name of output experiments file"
        }
    """,
    process_includes=True,
)


class Script:
    """A class to encapsulate the script."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        # Create the parser
        usage = "dials.apply_mask models.expt input.mask=pixels.mask"
        self.parser = ArgumentParser(
            usage=usage, epilog=help_message, phil=phil_scope, read_experiments=True
        )

    def run(self, args=None):
        """Run the script."""
        from dials.util import Sorry
        from dials.util.options import flatten_experiments

        # Parse the command line arguments
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        experiments = flatten_experiments(params.input.experiments)

        # Check that an experiment list and at least one mask file have been provided
        if not (experiments and params.input.mask):
            self.parser.print_help()
            return

        # Check number of experiments
        n_expts = len(experiments)
        n_masks = len(params.input.mask)
        if n_expts != n_masks:
            raise Sorry(
                "The number of masks provided must match the number of imagesets "
                "(sequences).\n"
                "You have provided an experiment list containing {} imageset(s).\n"
                "You have provided {} mask file(s).".format(n_expts, n_masks)
            )

        # Get the imageset
        imagesets = experiments.imagesets()

        for i, imageset in enumerate(imagesets):
            # Set the lookup
            with open(params.input.mask[i], "rb") as f:
                mask = pickle.load(f, encoding="bytes")
            imageset.external_lookup.mask.filename = params.input.mask[i]
            imageset.external_lookup.mask.data = ImageBool(mask)

        # Dump the experiments
        print(f"Writing experiments to {params.output.experiments}")
        experiments.as_file(filename=params.output.experiments)


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

import os

""" Test the situation that led to https://github.com/dials/dials/issues/423.
In that case instantiating a Refiner for an experiment list with an I23
detector model caused the panel origins to move before any refinement took
place. This occurred because for the input models.expt the root frame for
the hierarchical detector is on source side of the laboratory frame origin, not
on the detector side. Prior to the fix this resulted in incorrect calculation
of the offsets of all panels from the root frame.
"""


def test_run(dials_regression):
    from dxtbx.model.experiment_list import ExperimentListFactory
    from libtbx import phil

    from dials.algorithms.refinement import RefinerFactory
    from dials.algorithms.refinement.refiner import phil_scope
    from dials.array_family import flex

    data_dir = os.path.join(dials_regression, "refinement_test_data", "dials-423")
    exp_file = os.path.join(data_dir, "experiments.json")
    ref_file = os.path.join(data_dir, "subset.pickle")

    reflections = flex.reflection_table.from_file(ref_file)
    experiments = ExperimentListFactory.from_json_file(exp_file, check_format=False)
    """Test that the detector remains similar after refiner construction"""

    params = phil_scope.fetch(source=phil.parse("")).extract()

    # disable outlier rejection for speed of refiner construction
    params.refinement.reflections.outlier.algorithm = "null"

    refiner = RefinerFactory.from_parameters_data_experiments(
        params, reflections, experiments
    )

    d1 = experiments[0].detector
    d2 = refiner.get_experiments()[0].detector

    assert d1.is_similar_to(d2)



from __future__ import annotations

import collections

from dials.util import tabulate
from dials_algorithms_integration_ext import *  # noqa: F403; lgtm
from dials_algorithms_integration_kapton_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "Corrections",
    "CorrectionsMulti",
    "lp_correction",
    "qe_correction",
    "Result",
)

Result = collections.namedtuple(
    "Result",
    "index, reflections, data, read_time, extract_time, process_time, total_time",
)
#        :param index: The processing job index
#        :param reflections: The processed reflections
#        :param data: Other processed data


class TimingInfo:
    """
    A class to contain timing info.
    """

    def __init__(self):
        self.read = 0
        self.extract = 0
        self.initialize = 0
        self.process = 0
        self.finalize = 0
        self.total = 0
        self.user = 0

    def __str__(self):
        """Convert to string."""
        rows = [
            [description, f"{value:.2f} seconds"]
            for description, value in (
                ["Read time", self.read],
                ["Extract time", self.extract],
                ["Pre-process time", self.initialize],
                ["Process time", self.process],
                ["Post-process time", self.finalize],
                ["Total time", self.total],
                ["User time", self.user],
            )
            if value
        ]
        return tabulate(rows)

    def __add__(self, other):
        if not isinstance(other, TimingInfo):
            raise TypeError(
                f"unsupported addition: expected type TimingInfo, not {other!r}"
            )
        new_timing = TimingInfo()
        new_timing.read = self.read + other.read
        new_timing.extract = self.extract + other.extract
        new_timing.initialize = self.initialize + other.initialize
        new_timing.process = self.process + other.process
        new_timing.finalize = self.finalize + other.finalize
        new_timing.total = self.total + other.total
        new_timing.user = self.user + other.user
        return new_timing



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

from __future__ import annotations

import dials.util

help_message = """

This program is used to view the reflections with debugging purposes.
This program does not perform any calculation ... just visualizations

Example for invoking from CLI:

dials.reflection_viewer observations.refl
"""


class Script:
    """The debugging visualization program."""

    def __init__(self):
        """Initialise the script."""
        from dials.util.options import ArgumentParser

        usage = "dials.reflection_viewer [options] experiment.expt"

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage, epilog=help_message, read_reflections=True
        )

    def run(self, args=None):

        from dials.util.options import flatten_reflections
        from dials.viewer.viewer_interface import extract_n_show

        # Parse the command line
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        table = flatten_reflections(params.input.reflections)
        if len(table) == 0:
            self.parser.print_help()
            return

        extract_n_show(table[0])


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

from math import pi, sqrt

from dials.array_family import flex  # noqa: F401;
from dials_algorithms_profile_model_ellipsoid_ext import *  # noqa: F401, F403;


def mosaicity_from_eigen_decomposition(eigen_values):
    return (
        sqrt(eigen_values[0]) * 180.0 / pi,
        sqrt(eigen_values[1]) * 180.0 / pi,
        sqrt(eigen_values[2]) * 180.0 / pi,
    )



"""
Test derivatives typed up in dials_regression/doc/notes/prediction/stills_prediction_nave3.pdf
"""


from __future__ import annotations

import pytest

from cctbx.sgtbx import space_group, space_group_symbols
from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse
from scitbx import matrix

from dials.algorithms.refinement.parameterisation.beam_parameters import (
    BeamParameterisation,
)
from dials.algorithms.refinement.parameterisation.crystal_parameters import (
    CrystalOrientationParameterisation,
    CrystalUnitCellParameterisation,
)
from dials.algorithms.refinement.parameterisation.detector_parameters import (
    DetectorParameterisationSinglePanel,
)
from dials.algorithms.refinement.parameterisation.prediction_parameters_stills import (
    StillsPredictionParameterisation,
)
from dials.algorithms.refinement.prediction.managed_predictors import ScansRayPredictor
from dials.algorithms.spot_prediction import IndexGenerator, StillsReflectionPredictor
from dials.array_family import flex

from . import geometry_phil
from .setup_geometry import Extract


class Predictor:
    def __init__(self, experiments):

        self._experiment = experiments[0]
        self.update()

    def update(self):
        """Build predictor objects for the current geometry of each Experiment"""

        self._predictor = StillsReflectionPredictor(
            self._experiment, spherical_relp=True
        )
        self._UB = matrix.sqr(self._experiment.crystal.get_U()) * matrix.sqr(
            self._experiment.crystal.get_B()
        )

    def predict(self, reflections):
        """
        Predict for all reflections
        """

        self._predictor.for_reflection_table(reflections, self._UB)
        return reflections


# Simplied equivalent of a PredictionParameterisation class for this test
class AnalyticalGradients:
    """A class to implement the analytical gradient calculation in the document"""

    def __init__(
        self,
        experiments,
        detector_parameterisation,
        beam_parameterisation,
        xl_orientation_parameterisation,
        xl_unit_cell_parameterisation,
    ):

        # References to the underlying models from the first experiment
        self.experiment = experiments[0]
        self.beam = self.experiment.beam
        self.crystal = self.experiment.crystal

        # Keep references to all parameterisations
        self.detector_parameterisation = detector_parameterisation
        self.beam_parameterisation = beam_parameterisation
        self.xl_orientation_parameterisation = xl_orientation_parameterisation
        self.xl_unit_cell_parameterisation = xl_unit_cell_parameterisation

        # Keep quantities of interest from the model
        self.U = matrix.sqr(self.crystal.get_U())
        self.B = matrix.sqr(self.crystal.get_B())
        self.UB = self.U * self.B
        self.s0 = matrix.col(self.beam.get_s0())
        self.s0len = self.s0.length()
        self.s0len2 = self.s0.length_sq()
        self.us0 = self.s0.normalize()

    def get_beam_gradients(self, reflections):

        ds0_dbeam_p = self.beam_parameterisation.get_ds_dp()
        p_names = self.beam_parameterisation.get_param_names()

        n = len(reflections)
        U = flex.mat3_double(n, self.U)
        B = flex.mat3_double(n, self.B)
        UB = U * B

        # q is the reciprocal lattice vector, in the lab frame
        h = reflections["miller_index"].as_vec3_double()
        q = UB * h
        qlen2 = q.dot(q)

        q_s0 = q + self.s0
        s1 = reflections["s1"]
        ss = qlen2 + 2 * q.dot(self.s0) + self.s0len2
        assert (ss > 0.0).all_eq(True)
        s = flex.sqrt(ss)
        sss = s * ss
        inv_s = 1.0 / s
        inv_sss = 1.0 / sss

        # check equation 10
        tmp = self.s0len * (q_s0) / s
        for a, b in zip(s1, tmp):
            assert a == pytest.approx(b, abs=1e-7)

        ds1_dp = {}

        # loop through the parameters
        for name, der in zip(p_names, ds0_dbeam_p):

            # term1
            term1 = self.us0.dot(der) * q_s0 + self.s0len * (der)
            term1 = term1 * inv_s

            # term2
            term2 = self.s0len * q_s0 * q_s0.dot(der)
            term2 = term2 * inv_sss

            name = "Beam1" + name  # XXXX Hack to get matching keys
            ds1_dp[name] = {"ds1": (term1 - term2)}

        return ds1_dp

    def get_crystal_orientation_gradients(self, reflections):

        # get derivatives of the U matrix wrt the parameters
        dU_dxlo_p = self.xl_orientation_parameterisation.get_ds_dp()
        p_names = self.xl_orientation_parameterisation.get_param_names()

        n = len(reflections)
        U = flex.mat3_double(n, self.U)
        B = flex.mat3_double(n, self.B)
        UB = U * B

        # q is the reciprocal lattice vector, in the lab frame
        h = reflections["miller_index"].as_vec3_double()
        q = UB * h
        qlen2 = q.dot(q)

        q_s0 = q + self.s0
        ss = qlen2 + 2 * q.dot(self.s0) + self.s0len2
        assert (ss > 0.0).all_eq(True)
        s = flex.sqrt(ss)
        sss = s * ss
        inv_s = 1.0 / s
        inv_sss = 1.0 / sss

        ds1_dp = {}

        # loop through the parameters
        for name, der in zip(p_names, dU_dxlo_p):

            # calculate the derivative of q for this parameter
            dq = flex.mat3_double(n, der.elems) * B * h

            # term1
            term1 = self.s0len * dq
            term1 = term1 * inv_s

            # term2
            term2 = self.s0len * q_s0 * q_s0.dot(dq)
            term2 = term2 * inv_sss

            name = "Crystal1" + name  # XXXX Hack to get matching keys
            ds1_dp[name] = {"ds1": (term1 - term2)}

        return ds1_dp

    def get_crystal_unit_cell_gradients(self, reflections):

        # get derivatives of the B matrix wrt the parameters
        dB_dxluc_p = self.xl_unit_cell_parameterisation.get_ds_dp()
        p_names = self.xl_unit_cell_parameterisation.get_param_names()

        n = len(reflections)
        U = flex.mat3_double(n, self.U)
        B = flex.mat3_double(n, self.B)
        UB = U * B

        # q is the reciprocal lattice vector, in the lab frame
        h = reflections["miller_index"].as_vec3_double()
        q = UB * h
        qlen2 = q.dot(q)

        q_s0 = q + self.s0
        ss = qlen2 + 2 * q.dot(self.s0) + self.s0len2
        assert (ss > 0.0).all_eq(True)
        s = flex.sqrt(ss)
        sss = s * ss
        inv_s = 1.0 / s
        inv_sss = 1.0 / sss

        ds1_dp = {}

        # loop through the parameters
        for name, der in zip(p_names, dB_dxluc_p):

            # calculate the derivative of q for this parameter
            dq = U * flex.mat3_double(n, der.elems) * h

            # term1
            term1 = self.s0len * dq
            term1 = term1 * inv_s

            # term2
            term2 = self.s0len * q_s0 * q_s0.dot(dq)
            term2 = term2 * inv_sss

            name = "Crystal1" + name  # XXXX Hack to get matching keys
            ds1_dp[name] = {"ds1": (term1 - term2)}

        return ds1_dp


def test():
    # Build models, with a larger crystal than default in order to get plenty of
    # reflections on the 'still' image
    overrides = """
  geometry.parameters.crystal.a.length.range=40 50;
  geometry.parameters.crystal.b.length.range=40 50;
  geometry.parameters.crystal.c.length.range=40 50;
  geometry.parameters.random_seed = 42"""

    master_phil = parse(geometry_phil)

    models = Extract(master_phil, overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 3 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 1),
        exposure_times=0.1,
        oscillation=(0, 3.0),
        epochs=list(range(1)),
        deg=True,
    )
    sequence_range = myscan.get_oscillation_range(deg=False)

    # Create parameterisations of these models
    det_param = DetectorParameterisationSinglePanel(mydetector)
    s0_param = BeamParameterisation(mybeam, mygonio)
    xlo_param = CrystalOrientationParameterisation(mycrystal)
    xluc_param = CrystalUnitCellParameterisation(mycrystal)

    # Create a scans ExperimentList, only for generating reflections
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    # Create a stills ExperimentList
    stills_experiments = ExperimentList()
    stills_experiments.append(
        Experiment(beam=mybeam, detector=mydetector, crystal=mycrystal, imageset=None)
    )

    # Generate rays - only to work out which hkls are predicted
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    resolution = 2.0
    index_generator = IndexGenerator(
        mycrystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()
    rays = ray_predictor(indices)

    # Make a standard reflection_table and copy in the ray data
    reflections = flex.reflection_table.empty_standard(len(rays))
    reflections.update(rays)

    # Build a standard prediction parameterisation for the stills experiment to do
    # FD calculation (not used for its analytical gradients)
    pred_param = StillsPredictionParameterisation(
        stills_experiments,
        detector_parameterisations=[det_param],
        beam_parameterisations=[s0_param],
        xl_orientation_parameterisations=[xlo_param],
        xl_unit_cell_parameterisations=[xluc_param],
    )

    # Make a managed SphericalRelpStillsReflectionPredictor reflection predictor
    # for the first (only) experiment
    ref_predictor = Predictor(stills_experiments)

    # Predict these reflections in place. Must do this ahead of calculating
    # the analytical gradients so quantities like s1 are correct
    ref_predictor.update()
    ref_predictor.predict(reflections)

    # calculate analytical gradients
    ag = AnalyticalGradients(
        stills_experiments,
        detector_parameterisation=det_param,
        beam_parameterisation=s0_param,
        xl_orientation_parameterisation=xlo_param,
        xl_unit_cell_parameterisation=xluc_param,
    )
    an_grads = ag.get_beam_gradients(reflections)
    an_grads.update(ag.get_crystal_orientation_gradients(reflections))
    an_grads.update(ag.get_crystal_unit_cell_gradients(reflections))

    # get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    fd_grads = []
    p_names = pred_param.get_param_names()
    for i, delta in enumerate(deltas):

        # save parameter value
        val = p_vals[i]

        # calc reverse state
        p_vals[i] -= delta / 2.0
        pred_param.set_param_vals(p_vals)

        ref_predictor.update()
        ref_predictor.predict(reflections)

        x, y, _ = reflections["xyzcal.mm"].deep_copy().parts()
        s1 = reflections["s1"].deep_copy()
        rev_state = s1

        # calc forward state
        p_vals[i] += delta
        pred_param.set_param_vals(p_vals)

        ref_predictor.update()
        ref_predictor.predict(reflections)

        x, y, _ = reflections["xyzcal.mm"].deep_copy().parts()
        s1 = reflections["s1"].deep_copy()
        fwd_state = s1

        # reset parameter to saved value
        p_vals[i] = val

        # finite difference - currently for s1 only
        fd = fwd_state - rev_state
        inv_delta = 1.0 / delta
        s1_grads = fd * inv_delta

        # store gradients
        fd_grads.append({"name": p_names[i], "ds1": s1_grads})

    # return to the initial state
    pred_param.set_param_vals(p_vals)

    for i, fd_grad in enumerate(fd_grads):

        ## compare FD with analytical calculations
        print(f"\n\nParameter {i}: {fd_grad['name']}")

        print("d[s1]/dp for the first reflection")
        print("finite diff", fd_grad["ds1"][0])
        try:
            an_grad = an_grads[fd_grad["name"]]
        except KeyError:
            continue

        print("checking analytical vs finite difference gradients for s1")
        for a, b in zip(fd_grad["ds1"], an_grad["ds1"]):
            assert a == pytest.approx(b, abs=1e-7)



from __future__ import annotations

import iotbx.phil

import dials.util
from dials.array_family import flex
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

help_message = """
Augment spot list with additional information - for example number of pixels
in peak region etc."""

phil_scope = iotbx.phil.parse(
    """
output {
  reflections = stronger.refl
    .type = path
}
find_max = False
  .type = bool
"""
)


def add_max_pixels_to_reflections(reflections):
    """Iterate through reflections, find max pixel in each shoebox and max
    valid, add columns thereof"""

    from dials.algorithms.shoebox import MaskCode

    good = MaskCode.Foreground | MaskCode.Valid

    max_pixel = flex.double(reflections.size(), 0.0)
    max_valid = flex.double(reflections.size(), 0.0)

    shoeboxes = reflections["shoebox"]

    for j, s in enumerate(shoeboxes):
        max_pixel[j] = flex.max(s.data)
        valid = s.mask.as_1d() == good
        max_valid[j] = flex.max(s.data.as_1d().select(valid))

    reflections["max_pixel"] = max_pixel
    reflections["max_valid"] = max_valid


def add_resolution_to_reflections(reflections, experiments):
    """Add column d to reflection list"""

    # will assume everything from the first detector at the moment - clearly this
    # could be incorrect, will have to do something a little smarter, later

    if "imageset_id" not in reflections:
        reflections["imageset_id"] = reflections["id"]

    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    reflections["d"] = 1 / reflections["rlp"].norms()


def augment_reflections(reflections, params, experiments=None):
    """Add extra columns of derived data."""

    from dials.algorithms.shoebox import MaskCode

    good = MaskCode.Foreground | MaskCode.Valid

    if params.find_max:
        add_max_pixels_to_reflections(reflections)

    x0, x1, y0, y1, z0, z1 = reflections["bbox"].parts()

    dx = x1 - x0
    dy = y1 - y0

    # compute signal pixels in each shoebox as an array
    n_signal = reflections["shoebox"].count_mask_values(good)

    reflections["dx"] = dx
    reflections["dy"] = dy
    reflections["n_signal"] = n_signal

    if experiments:
        add_resolution_to_reflections(reflections, experiments)

    return reflections


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util import Sorry

    usage = "dials.augment_spots [options] [models.expt] strong.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(reflections) != 1:
        parser.print_help()
        return
    if len(experiments) > 1:
        raise Sorry("0, 1 experiments required")

    stronger = augment_reflections(reflections[0], params, experiments=experiments)
    stronger.as_file(params.output.reflections)


if __name__ == "__main__":
    run()



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1

"""
Visualise the reference profiles from integration.

Examples::

  dials.integrate refined.expt refined.refl debug.reference.output=True
  dials.reference_profile_viewer reference_profiles.pickle
"""

from __future__ import annotations

import copy
import os
import pickle

import matplotlib
import numpy
import wx

matplotlib.use("WXAgg")

from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigCanvas
from matplotlib.backends.backend_wxagg import (
    NavigationToolbar2WxAgg as NavigationToolbar,
)
from matplotlib.figure import Figure

import dials.util
import dials.util.log
from dials.util.options import ArgumentParser


class ProfilesFrame(wx.Frame):
    """The main frame of the application"""

    title = "Reference profile viewer"

    def __init__(self, profiles):
        wx.Frame.__init__(self, None, -1, self.title)

        self.profiles = profiles
        self.data = self.profiles.get_profiles(experiment=0, block=0)

        self.create_menu()
        self.create_status_bar()
        self.create_main_panel()
        self.draw_figure()

    def create_menu(self):
        self.menubar = wx.MenuBar()

        menu_file = wx.Menu()
        m_expt = menu_file.Append(-1, "&Save plot\tCtrl-S", "Save plot to file")
        self.Bind(wx.EVT_MENU, self.on_save_plot, m_expt)
        menu_file.AppendSeparator()
        m_exit = menu_file.Append(-1, "E&xit\tCtrl-X", "Exit")
        self.Bind(wx.EVT_MENU, self.on_exit, m_exit)

        menu_help = wx.Menu()
        m_about = menu_help.Append(-1, "&About\tF1", "About the program")
        self.Bind(wx.EVT_MENU, self.on_about, m_about)

        self.menubar.Append(menu_file, "&File")
        self.menubar.Append(menu_help, "&Help")
        self.SetMenuBar(self.menubar)

    def create_main_panel(self):
        """Creates the main panel with all the controls on it:
        * mpl canvas
        * mpl navigation toolbar
        * Control panel for interaction
        """
        self.panel = wx.Panel(self)

        # Create the mpl Figure and FigCanvas objects.
        # 7x5 inches, 100 dots-per-inch
        self.dpi = 100
        self.fig = Figure((7.0, 5.0), dpi=self.dpi)
        self.canvas = FigCanvas(self.panel, -1, self.fig)

        self.set_axes()

        self.expt_selection_label = wx.StaticText(self.panel, -1, "Experiment ID: ")
        self.expt_selection = wx.SpinCtrl(
            self.panel, -1, "0", max=self.profiles.get_n_experiments() - 1
        )
        self.Bind(wx.EVT_SPINCTRL, self.on_spin_expt, self.expt_selection)

        self.block_selection_label = wx.StaticText(self.panel, -1, "Block: ")
        self.block_selection = wx.SpinCtrl(
            self.panel, -1, "0", max=self.profiles.get_n_blocks(experiment=0) - 1
        )
        self.Bind(wx.EVT_SPINCTRL, self.on_spin_block, self.block_selection)

        self.mask_checkbox = wx.CheckBox(self.panel, -1, "Mask")
        self.mask_checkbox.SetValue(True)
        self.Bind(wx.EVT_CHECKBOX, self.redraw_on_event, self.mask_checkbox)

        self.cmap_choice_label = wx.StaticText(self.panel, -1, "Colourmap: ")
        self.cmap_choice = wx.Choice(
            self.panel, -1, choices=["viridis", "viridis_r", "gray", "gray_r"]
        )
        self.cmap_choice.SetSelection(0)
        self.Bind(wx.EVT_CHOICE, self.redraw_on_event, self.cmap_choice)

        # Create the navigation toolbar, tied to the canvas
        self.toolbar = NavigationToolbar(self.canvas)

        # Layout with box sizers
        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.vbox.Add(self.toolbar, 0, wx.EXPAND)
        self.vbox.AddSpacer(10)

        self.hbox = wx.BoxSizer(wx.HORIZONTAL)
        flags = wx.ALIGN_LEFT | wx.ALL | wx.ALIGN_CENTER_VERTICAL
        self.hbox.Add(self.expt_selection_label, 0, flag=flags)
        self.hbox.Add(self.expt_selection, 0, border=3, flag=flags)
        self.hbox.Add(self.block_selection_label, 0, flag=flags)
        self.hbox.Add(self.block_selection, 0, border=3, flag=flags)
        self.hbox.Add(self.mask_checkbox, 0, border=3, flag=flags)
        self.hbox.Add(self.cmap_choice_label, 0, flag=flags)
        self.hbox.Add(self.cmap_choice, 0, border=3, flag=flags)

        self.vbox.Add(self.hbox, 0, flag=wx.ALIGN_LEFT | wx.TOP)

        self.panel.SetSizer(self.vbox)
        self.vbox.Fit(self)

    def set_axes(self):
        subplots = [e["subplot"] for e in self.data]
        r, c = list(zip(*subplots))
        nrows = max(r) + 1
        ncols = max(c) + 1
        self.axes = self.fig.subplots(nrows, ncols, sharex=True, sharey=True)

    def create_status_bar(self):
        self.statusbar = self.CreateStatusBar()

    def draw_figure(self):
        """Redraws the figure"""

        final_row_index = self.axes.shape[0] - 1
        for profile in self.data:
            subplot = profile["subplot"]
            ax = self.axes[subplot]
            ax.clear()

            # For now, let's just sum down the first axis. Profiles are stored
            # in ez, ey, ex order, where ex is orthogonal to s1 and s0, ey is
            # orthogonal to s1 and ex, and ez is the axis that is dependent on
            # the direction through the Ewald sphere
            vals2D = profile["data"].sum(axis=0)
            cmap = copy.copy(
                matplotlib.cm.get_cmap(
                    self.cmap_choice.GetString(self.cmap_choice.GetSelection())
                )
            )

            # If any X, Y position is masked down the summed Z stack then mask
            # it in the final image.
            if self.mask_checkbox.IsChecked():
                mask2D = (profile["mask"] - 1).sum(axis=0)
                mask2D = mask2D != 0
                vals2D[mask2D] = numpy.nan
            cmap.set_bad(color="red")
            ax.imshow(vals2D, cmap=cmap)

            if subplot[0] == final_row_index:
                ax.set_xlabel(f"X (px): {profile['coord'][0]:.1f}")
            if subplot[1] == 0:
                ax.set_ylabel(f"Y (px): {profile['coord'][1]:.1f}")

        self.fig.suptitle(f"Block Z (im): {profile['coord'][2]:.1f}")
        self.canvas.draw()

    def on_spin_expt(self, event):
        self.expt_selection.Disable()
        exp_id = self.expt_selection.GetValue()
        self.data = self.profiles.get_profiles(experiment=exp_id, block=0)
        self.block_selection.SetValue(0)
        self.block_selection.SetMax(self.profiles.get_n_blocks(exp_id) - 1)
        self.draw_figure()
        self.expt_selection.Enable()

    def on_spin_block(self, event):
        self.block_selection.Disable()
        self.data = self.profiles.get_profiles(
            experiment=self.expt_selection.GetValue(),
            block=self.block_selection.GetValue(),
        )
        self.draw_figure()
        self.block_selection.Enable()

    def redraw_on_event(self, event):
        self.draw_figure()

    def on_save_plot(self, event):
        file_choices = "PNG (*.png)|*.png"

        dlg = wx.FileDialog(
            self,
            message="Save plot as...",
            defaultDir=os.getcwd(),
            defaultFile="reference_profiles.png",
            wildcard=file_choices,
            style=wx.SAVE,
        )

        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
            self.canvas.print_figure(path, dpi=self.dpi)
            self.flash_status_message("Saved to %s" % path)

    def on_exit(self, event):
        self.Destroy()

    def on_about(self, event):
        msg = """A reference profile viewer for DIALS:

         * The reference profiles for the selected block of
           images (Z) are displayed along with their X, Y
           pixel positions
         * The display sums slices of the Kabsch space shoebox
           down the e3 direction, hence the view is the
           integration of the profile as it passes through
           the Ewald sphere
         * If any of the slices contains a masked pixel, the
           equivalent pixel in the summed image is also
           showed as masked
         * Save the plot to a file using the File menu
        """
        dlg = wx.MessageDialog(self, msg, "About", wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def flash_status_message(self, msg, flash_len_ms=1500):
        self.statusbar.SetStatusText(msg)
        self.timeroff = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_flash_status_off, self.timeroff)
        self.timeroff.Start(flash_len_ms, oneShot=True)

    def on_flash_status_off(self, event):
        self.statusbar.SetStatusText("")


class ProfileStore:
    def __init__(self, filename):
        with open(filename, "rb") as f:
            data = pickle.load(f)

        self._experiment_data = []
        for d in data:
            self._experiment_data.append(self._process_experiment(d))

    def _process_experiment(self, data):
        # Reorganise data by Z block
        z_blocks = {e["coord"][2] for e in data}
        z_blocks = dict.fromkeys(z_blocks)
        for model in data:
            z_coord = model["coord"][2]
            if z_blocks[z_coord] is None:
                z_blocks[z_coord] = [
                    model,
                ]
            else:
                z_blocks[z_coord].append(model)

        # Process within Z blocks
        for k, v in z_blocks.items():
            z_blocks[k] = self._process_block(v)

        # Convert to list
        keys = sorted(z_blocks.keys())
        return [z_blocks[k] for k in keys]

    @staticmethod
    def _process_block(block):
        x_coords = sorted({e["coord"][0] for e in block})
        y_coords = sorted({e["coord"][1] for e in block})

        for profile in block:
            x, y, _ = profile["coord"]
            i = x_coords.index(x)
            j = y_coords.index(y)
            profile["subplot"] = (j, i)  # row (Y), then column (X)
            profile["data"] = profile["data"].as_numpy_array()
            profile["mask"] = profile["mask"].as_numpy_array()

        return block

    def get_n_experiments(self):
        return len(self._experiment_data)

    def get_n_blocks(self, experiment):
        return len(self._experiment_data[experiment])

    def get_profiles(self, experiment, block):
        return self._experiment_data[experiment][block]


@dials.util.show_mail_on_error()
def run():
    dials.util.log.print_banner()
    usage = "dials.reference_profile_viewer [options] reference_profiles.pickle"
    parser = ArgumentParser(
        usage=usage,
        read_reflections=False,
        read_experiments=False,
        check_format=False,
        epilog=__doc__,
    )

    params, options, args = parser.parse_args(
        show_diff_phil=False, return_unhandled=True
    )
    if len(args) != 1:
        parser.print_help()
        exit(0)
    filename = args[0]
    assert os.path.isfile(filename), filename

    # Load data
    data = ProfileStore(filename)

    # Start viewer
    show_reference_profile_viewer(data, params)


def show_reference_profile_viewer(data, params):
    app = wx.App()
    app.frame = ProfilesFrame(data)
    app.frame.Show()
    app.MainLoop()


if __name__ == "__main__":
    run()



from __future__ import annotations

import json
import logging
import sys

import libtbx.phil

from dials.util import detect_blanks, show_mail_handle_errors

logger = logging.getLogger("dials.detect_blanks")

phil_scope = libtbx.phil.parse(
    """\
phi_step = 2
  .type = float(value_min=0)
  .help = "Width of bins in degrees."
counts_fractional_loss = 0.1
  .type = float(value_min=0, value_max=1)
  .help = "Fractional loss (relative to the bin with the most counts) after "
          "which a bin is flagged as potentially containing blank images."
misigma_fractional_loss = 0.1
  .type = float(value_min=0, value_max=1)
  .help = "Fractional loss (relative to the bin with the highest misigma) after "
          "which a bin is flagged as potentially containing blank images."
output {
  json = blanks.json
    .type = path
  plot = False
    .type = bool
}
""",
    process_includes=True,
)


help_message = """\
"""


@show_mail_handle_errors()
def run(args=None):
    from dials.util import log
    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )

    usage = "dials.detect_blanks [options] models.expt observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 or len(reflections) == 0:
        parser.print_help()
        exit(0)

    # Configure the logging
    log.config(logfile="dials.detect_blanks.log")

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    reflections = reflections[0]

    imagesets = experiments.imagesets()

    if any(experiment.is_still() for experiment in experiments):
        sys.exit("dials.detect_blanks can only be used with rotation data")

    assert len(imagesets) == 1
    imageset = imagesets[0]
    scan = imageset.get_scan()

    integrated_sel = reflections.get_flags(reflections.flags.integrated)
    indexed_sel = reflections.get_flags(reflections.flags.indexed)
    centroid_outlier_sel = reflections.get_flags(reflections.flags.centroid_outlier)
    strong_sel = reflections.get_flags(reflections.flags.strong)
    indexed_sel &= ~centroid_outlier_sel

    logger.info(f"Analysis of {strong_sel.count(True)} strong reflections:")
    strong_results = detect_blanks.blank_counts_analysis(
        reflections.select(strong_sel),
        scan,
        phi_step=params.phi_step,
        fractional_loss=params.counts_fractional_loss,
    )
    for blank_start, blank_end in strong_results["blank_regions"]:
        logger.info(f"Potential blank images: {blank_start + 1} -> {blank_end}")

    indexed_results = None
    if indexed_sel.count(True) > 0:
        logger.info(f"Analysis of {indexed_sel.count(True)} indexed reflections:")
        indexed_results = detect_blanks.blank_counts_analysis(
            reflections.select(indexed_sel),
            scan,
            phi_step=params.phi_step,
            fractional_loss=params.counts_fractional_loss,
        )
        for blank_start, blank_end in indexed_results["blank_regions"]:
            logger.info(f"Potential blank images: {blank_start + 1} -> {blank_end}")

    integrated_results = None
    if integrated_sel.count(True) > 0:
        logger.info(f"Analysis of {integrated_sel.count(True)} integrated reflections:")
        integrated_results = detect_blanks.blank_integrated_analysis(
            reflections.select(integrated_sel),
            scan,
            phi_step=params.phi_step,
            fractional_loss=params.misigma_fractional_loss,
        )
        for blank_start, blank_end in integrated_results["blank_regions"]:
            logger.info(f"Potential blank images: {blank_start + 1} -> {blank_end}")

    d = {
        "strong": strong_results,
        "indexed": indexed_results,
        "integrated": integrated_results,
    }

    if params.output.json is not None:
        with open(params.output.json, "w") as fh:
            json.dump(d, fh)

    if params.output.plot:
        from matplotlib import pyplot

        plots = [(strong_results, "-")]
        if indexed_results:
            plots.append((indexed_results, "--"))
        if integrated_results:
            plots.append((integrated_results, ":"))

        for results, linestyle in plots:
            xs = results["data"][0]["x"]
            ys = results["data"][0]["y"]
            xmax = max(xs)
            ymax = max(ys)
            xs = [x / xmax for x in xs]
            ys = [y / ymax for y in ys]
            blanks = results["data"][0]["blank"]
            pyplot.plot(xs, ys, color="blue", linestyle=linestyle)
            pyplot.plot(
                *zip(*[(x, y) for x, y, blank in zip(xs, ys, blanks) if blank]),
                color="red",
                linestyle=linestyle,
            )
        pyplot.ylim(0)
        pyplot.show()
        pyplot.clf()


if __name__ == "__main__":
    run()



from __future__ import annotations

import os

import procrunner


def test(dials_regression, run_in_tmp_path):
    result = procrunner.run(
        [
            "dials.plot_scan_varying_model",
            os.path.join(
                dials_regression,
                "refinement_test_data",
                "multi_sweep_one_sample",
                "glucose_isomerase",
                "SWEEP1",
                "index",
                "sv_refined_experiments.json",
            ),
        ]
    )
    assert not result.returncode and not result.stderr



from __future__ import annotations

import logging
import random
import sys

import numpy as np

import iotbx.phil
from cctbx import sgtbx

from dials.algorithms.clustering.unit_cell import cluster_unit_cells
from dials.algorithms.symmetry.cosym import CosymAnalysis, extract_reference_intensities
from dials.algorithms.symmetry.cosym.observers import register_default_cosym_observers
from dials.array_family import flex
from dials.command_line.symmetry import (
    apply_change_of_basis_ops,
    change_of_basis_ops_to_minimum_cell,
    eliminate_sys_absent,
)
from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.exclude_images import get_selection_for_valid_image_ranges
from dials.util.filter_reflections import filtered_arrays_from_experiments_reflections
from dials.util.multi_dataset_handling import (
    assign_unique_identifiers,
    parse_multiple_datasets,
    select_datasets_on_identifiers,
    update_imageset_ids,
)
from dials.util.observer import Subject
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.cosym")

phil_scope = iotbx.phil.parse(
    """\
partiality_threshold = 0.4
  .type = float
  .help = "Use reflections with a partiality above the threshold."

unit_cell_clustering {
  threshold = 5000
    .type = float(value_min=0, allow_none=True)
    .help = 'Threshold value for the clustering'
  log = False
    .type = bool
    .help = 'Display the dendrogram with a log scale'
}

reference = None
    .type = path
    .help = "A file containing a reference set of intensities e.g. MTZ/cif, or a"
            "file from which a reference set of intensities can be calculated"
            "e.g. .pdb or .cif . The space group of the reference file will"
            "be used and if an indexing ambiguity is present, the input"
            "data will be reindexed to be consistent with the indexing mode of"
            "this reference file."
    .expert_level = 2

include scope dials.algorithms.symmetry.cosym.phil_scope

relative_length_tolerance = 0.05
  .type = float(value_min=0)

absolute_angle_tolerance = 2
  .type = float(value_min=0)

min_reflections = 10
  .type = int(value_min=1)
  .help = "The minimum number of reflections per experiment."

seed = 230
  .type = int(value_min=0)

output {
  suffix = "_reindexed"
    .type = str
  log = dials.cosym.log
    .type = str
  experiments = "symmetrized.expt"
    .type = path
  reflections = "symmetrized.refl"
    .type = path
  json = dials.cosym.json
    .type = path
  html = dials.cosym.html
    .type = path
}
""",
    process_includes=True,
)


class cosym(Subject):
    def __init__(self, experiments, reflections, params=None):
        super().__init__(events=["run_cosym", "performed_unit_cell_clustering"])
        if params is None:
            params = phil_scope.extract()
        self.params = params

        reference_intensities = None
        if self.params.reference:
            reference_intensities, space_group_info = extract_reference_intensities(
                params
            )
            if self.params.space_group and (
                self.params.space_group.type().number()
                != space_group_info.type().number()
            ):
                # N.B. space group phil options are actually space_group_info objects
                raise ValueError(
                    f"Input space group ({self.params.space_group}) does not match space group from reference file ({space_group_info})"
                )
            logger.info(f"Using space group {space_group_info} from reference")
            self.params.space_group = space_group_info

        self._reflections = []
        for refl, expt in zip(reflections, experiments):
            sel = get_selection_for_valid_image_ranges(refl, expt)
            self._reflections.append(refl.select(sel))

        self._experiments, self._reflections = self._filter_min_reflections(
            experiments, self._reflections
        )
        self.ids_to_identifiers_map = {}
        for table in self._reflections:
            self.ids_to_identifiers_map.update(table.experiment_identifiers())
        self.identifiers_to_ids_map = {
            value: key for key, value in self.ids_to_identifiers_map.items()
        }

        if len(self._experiments) > 1 and self.params.unit_cell_clustering.threshold:
            # perform unit cell clustering
            identifiers = self._unit_cell_clustering(self._experiments)
            if len(identifiers) < len(self._experiments):
                logger.info(
                    "Selecting subset of %i datasets for cosym analysis: %s",
                    len(identifiers),
                    str(identifiers),
                )
                self._experiments, self._reflections = select_datasets_on_identifiers(
                    self._experiments, self._reflections, use_datasets=identifiers
                )

        # Map experiments and reflections to minimum cell
        cb_ops = change_of_basis_ops_to_minimum_cell(
            self._experiments,
            params.lattice_symmetry_max_delta,
            params.relative_length_tolerance,
            params.absolute_angle_tolerance,
        )
        exclude = [
            expt.identifier
            for expt, cb_op in zip(self._experiments, cb_ops)
            if not cb_op
        ]
        if len(exclude):
            logger.info(
                f"Rejecting {len(exclude)} datasets from cosym analysis "
                f"(couldn't determine consistent cb_op to minimum cell):\n"
                f"{exclude}",
            )
            self._experiments, self._reflections = select_datasets_on_identifiers(
                self._experiments, self._reflections, exclude_datasets=exclude
            )
            cb_ops = list(filter(None, cb_ops))

        # Eliminate reflections that are systematically absent due to centring
        # of the lattice, otherwise they would lead to non-integer miller indices
        # when reindexing to a primitive setting
        self._reflections = eliminate_sys_absent(self._experiments, self._reflections)

        self._experiments, self._reflections = apply_change_of_basis_ops(
            self._experiments, self._reflections, cb_ops
        )

        # transform models into miller arrays
        datasets = filtered_arrays_from_experiments_reflections(
            self.experiments,
            self.reflections,
            outlier_rejection_after_filter=False,
            partiality_threshold=params.partiality_threshold,
        )

        datasets = [
            ma.as_non_anomalous_array().merge_equivalents().array() for ma in datasets
        ]
        if reference_intensities:
            datasets.append(reference_intensities)
            self.cosym_analysis = CosymAnalysis(
                datasets, self.params, seed_dataset=len(datasets) - 1
            )
        else:
            self.cosym_analysis = CosymAnalysis(datasets, self.params)

    @property
    def experiments(self):
        """Return the experiment list."""
        return self._experiments

    @property
    def reflections(self):
        """Return the list of reflection tables."""
        return self._reflections

    @Subject.notify_event(event="run_cosym")
    def run(self):
        self.cosym_analysis.run()
        reindexing_ops = self.cosym_analysis.reindexing_ops
        datasets_ = list(set(self.cosym_analysis.dataset_ids))

        # Log reindexing operators
        logger.info("Reindexing operators:")
        for cb_op in set(reindexing_ops):
            datasets = [d for d, o in zip(datasets_, reindexing_ops) if o == cb_op]
            logger.info(f"{cb_op}: {datasets}")

        self._apply_reindexing_operators(
            reindexing_ops, subgroup=self.cosym_analysis.best_subgroup
        )

    def export(self):
        """Output the datafiles for cosym.

        This includes the cosym.json, reflections and experiments files."""

        reindexed_reflections = flex.reflection_table()
        self._reflections = update_imageset_ids(self._experiments, self._reflections)
        for refl in self._reflections:
            reindexed_reflections.extend(refl)
        reindexed_reflections.reset_ids()

        logger.info(
            "Saving reindexed experiments to %s", self.params.output.experiments
        )
        self._experiments.as_file(self.params.output.experiments)
        logger.info(
            "Saving reindexed reflections to %s", self.params.output.reflections
        )
        reindexed_reflections.as_file(self.params.output.reflections)

    def _apply_reindexing_operators(self, reindexing_ops, subgroup=None):
        """Apply the reindexing operators to the reflections and experiments."""
        if self.params.reference:
            unique_ids = set(self.cosym_analysis.dataset_ids[:-1])
            reindexing_ops = reindexing_ops[:-1]
        else:
            unique_ids = set(self.cosym_analysis.dataset_ids)
        for cb_op, dataset_id in zip(reindexing_ops, unique_ids):
            cb_op = sgtbx.change_of_basis_op(cb_op)
            logger.debug(
                "Applying reindexing op %s to dataset %i", cb_op.as_xyz(), dataset_id
            )
            expt = self._experiments[dataset_id]
            refl = self._reflections[dataset_id]
            if subgroup is not None:
                cb_op = subgroup["cb_op_inp_best"] * cb_op
                expt.crystal = expt.crystal.change_basis(cb_op)
                expt.crystal.set_space_group(
                    subgroup["best_subsym"].space_group().build_derived_acentric_group()
                )
            else:
                expt.crystal = expt.crystal.change_basis(cb_op)
            expt.crystal.set_unit_cell(
                expt.crystal.get_space_group().average_unit_cell(
                    expt.crystal.get_unit_cell()
                )
            )
            refl["miller_index"] = cb_op.apply(refl["miller_index"])
        # Allow for the case where some datasets are filtered out.
        if len(reindexing_ops) < len(self._experiments):
            to_delete = [
                i for i in range(len(self._experiments)) if i not in unique_ids
            ]
            for idx in sorted(to_delete, reverse=True):
                logger.info(
                    f"Removing dataset {idx} as unable to determine reindexing operator"
                )
                del self._experiments[idx]
                del self._reflections[idx]

    def _filter_min_reflections(self, experiments, reflections):
        identifiers = []

        for expt, refl in zip(experiments, reflections):
            if len(refl) >= self.params.min_reflections:
                identifiers.append(expt.identifier)

        return select_datasets_on_identifiers(
            experiments, reflections, use_datasets=identifiers
        )

    @Subject.notify_event("performed_unit_cell_clustering")
    def _unit_cell_clustering(self, experiments):
        crystal_symmetries = [
            expt.crystal.get_crystal_symmetry() for expt in experiments
        ]
        # lattice ids used to label plots, so want numerical ids
        lattice_ids = [
            self.identifiers_to_ids_map[i] for i in experiments.identifiers()
        ]
        assert len(crystal_symmetries) > 1  # else clustering will be None
        # assert should never be tripped as len(experiments) are checked before
        # calling the _unit_cell_clustering function.
        clustering = cluster_unit_cells(
            crystal_symmetries,
            lattice_ids=lattice_ids,
        )
        self.unit_cell_clusters = clustering.clusters
        self.unit_cell_dendrogram = clustering.dendrogram

        logger.info(clustering)
        largest_cluster_lattice_ids = None
        for cluster in self.unit_cell_clusters:
            cluster_lattice_ids = cluster.lattice_ids
            if largest_cluster_lattice_ids is None:
                largest_cluster_lattice_ids = cluster_lattice_ids
            elif len(cluster_lattice_ids) > len(largest_cluster_lattice_ids):
                largest_cluster_lattice_ids = cluster_lattice_ids

        dataset_selection = largest_cluster_lattice_ids
        # now convert to actual identifiers for selection
        return [self.ids_to_identifiers_map[i] for i in dataset_selection]


help_message = """
This program implements the methods of `Gildea, R. J. & Winter, G. (2018).
Acta Cryst. D74, 405-410 <https://doi.org/10.1107/S2059798318002978>`_ for
determination of Patterson group symmetry from sparse multi-crystal data sets in
the presence of an indexing ambiguity.

The program takes as input a set of integrated experiments and reflections,
either in one file per experiment, or with all experiments combined in a single
models.expt and observations.refl file. It will perform analysis of the
symmetry elements present in the datasets and, if necessary, reindex experiments
and reflections as necessary to ensure that all output experiments and
reflections are indexed consistently.

Examples::

  dials.cosym models.expt observations.refl

  dials.cosym models.expt observations.refl space_group=I23

  dials.cosym models.expt observations.refl space_group=I23 lattice_group=I23
"""


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.cosym [options] models.expt observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options, args = parser.parse_args(
        args=args, show_diff_phil=False, return_unhandled=True
    )

    # Configure the logging
    log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    if params.seed is not None:
        flex.set_random_seed(params.seed)
        np.random.seed(params.seed)
        random.seed(params.seed)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    reflections = parse_multiple_datasets(reflections)
    if len(experiments) != len(reflections):
        raise Sorry(
            "Mismatched number of experiments and reflection tables found: %s & %s."
            % (len(experiments), len(reflections))
        )
    try:
        experiments, reflections = assign_unique_identifiers(experiments, reflections)
        cosym_instance = cosym(
            experiments=experiments, reflections=reflections, params=params
        )
    except ValueError as e:
        raise Sorry(e)

    if params.output.html or params.output.json:
        register_default_cosym_observers(cosym_instance)
    cosym_instance.run()
    cosym_instance.export()


if __name__ == "__main__":
    run()



"""Tests for dials.split_experiments when experiment ids are set"""

from __future__ import annotations

import procrunner
import pytest

from dxtbx.model import Beam, Experiment, ExperimentList
from dxtbx.serialize import load

from dials.array_family import flex


def generate_exp(wavelength=1):
    """Generate an experiment containing a beam with a given wavelength."""
    beam = Beam(direction=(0.0, 0.0, 1.0), wavelength=wavelength)
    exp = Experiment(beam=beam)
    return exp


@pytest.mark.parametrize("with_identifiers", ["True", "False"])
@pytest.mark.parametrize("option", ["chunk_size=2", 'chunk_sizes="2 2 1"'])
def test_split_chunk_sizes(tmpdir, option, with_identifiers):

    ids = list(range(0, 5))
    experiments = ExperimentList()
    reflections = flex.reflection_table()
    reflections["id"] = flex.int(ids)
    reflections["intensity"] = flex.double([(i + 1) * 100.0 for i in ids])

    for i in ids:
        exp = generate_exp()
        if with_identifiers:
            exp.identifier = str(i)
            reflections.experiment_identifiers()[i] = str(i)
        experiments.append(exp)

    experiments.as_json(tmpdir.join("tmp.expt").strpath)
    reflections.as_file(tmpdir.join("tmp.refl").strpath)

    result = procrunner.run(
        [
            "dials.split_experiments",
            tmpdir.join("tmp.expt").strpath,
            tmpdir.join("tmp.refl").strpath,
            option,
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    for j, n, intensities in zip(
        [0, 1, 2], [2, 2, 1], [[100.0, 200.0], [300.0, 400.0], [500.0]]
    ):
        assert tmpdir.join(f"split_{j}.refl").check()
        assert tmpdir.join(f"split_{j}.expt").check()
        expts = load.experiment_list(tmpdir.join(f"split_{j}.expt"), check_format=False)
        assert len(expts) == n
        refls = flex.reflection_table.from_file(tmpdir.join(f"split_{j}.refl"))
        assert list(set(refls["id"])) == list(range(0, n))
        assert list(refls["intensity"]) == intensities
        refls.assert_experiment_identifiers_are_consistent(expts)


def test_split_by_wavelength(tmpdir):
    """Test the split_by_wavelength option of dials.split_experiments"""
    experiments = ExperimentList()
    exp = generate_exp(wavelength=1.0)
    exp.identifier = "0"
    experiments.append(exp)
    exp = generate_exp(wavelength=0.5)
    exp.identifier = "1"
    experiments.append(exp)

    reflections = flex.reflection_table()
    reflections["id"] = flex.int([0, 1])
    reflections["intensity"] = flex.double([100.0, 200.0])
    reflections.experiment_identifiers()[0] = "0"
    reflections.experiment_identifiers()[1] = "1"

    experiments.as_json(tmpdir.join("tmp.expt").strpath)
    reflections.as_file(tmpdir.join("tmp.refl").strpath)

    result = procrunner.run(
        ["dials.split_experiments", "tmp.expt", "tmp.refl", "by_wavelength=True"],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr

    for i, (wl, ids, intensity) in enumerate(
        zip([0.5, 1.0], ["1", "0"], [200.0, 100.0])
    ):
        assert tmpdir.join("split_%d.expt" % i).check()
        assert tmpdir.join("split_%d.refl" % i).check()
        exp_single = load.experiment_list(
            tmpdir.join("split_%d.expt" % i).strpath, check_format=False
        )
        ref_single = flex.reflection_table.from_file(
            tmpdir.join("split_%d.refl" % i).strpath
        )
        assert exp_single[0].beam.get_wavelength() == wl
        assert exp_single[0].identifier == ids
        id_ = ref_single["id"][0]
        assert ref_single.experiment_identifiers()[id_] == ids
        assert list(ref_single["intensity"]) == [intensity]

    # Now test for successful error handling if no identifiers set.
    experiments[0].identifier = ""
    experiments[1].identifier = ""
    experiments.as_json(tmpdir.join("tmp.expt").strpath)
    result = procrunner.run(
        ["dials.split_experiments", "tmp.expt", "tmp.refl", "by_wavelength=True"],
        working_directory=tmpdir,
    )
    assert result.returncode == 1
    assert result.stderr.startswith(b"Sorry")

    experiments[0].identifier = "0"
    experiments[1].identifier = "1"
    del reflections.experiment_identifiers()[0]
    del reflections.experiment_identifiers()[1]
    experiments.as_json(tmpdir.join("tmp.expt").strpath)
    reflections.as_file(tmpdir.join("tmp.refl").strpath)
    result = procrunner.run(
        ["dials.split_experiments", "tmp.expt", "tmp.refl", "by_wavelength=True"],
        working_directory=tmpdir,
    )
    assert result.returncode == 1
    assert result.stderr.startswith(b"Sorry")



# LIBTBX_SET_DISPATCHER_NAME dials.background
# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1
"""Calculate background in resolution shells and optionally produce a plot.

Usage:
    dials.background image_*.cbf

    dials.background imported.expt

    dials.background imported.expt images=1,10 plot=background.png

"""

from __future__ import annotations

import logging
import math

import iotbx.phil
from libtbx.phil import parse
from scitbx import matrix

import dials.util.masking
from dials.algorithms.spot_finding.factory import SpotFinderFactory
from dials.algorithms.spot_finding.factory import phil_scope as spot_phil
from dials.array_family import flex
from dials.util import Sorry, log, show_mail_handle_errors
from dials.util.options import ArgumentParser, flatten_experiments
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.background")

phil_scope = iotbx.phil.parse(
    """\
n_bins = 100
  .type = int
n_checkpoints = None
  .type = int(value_min=1)
  .help = "Number of evenly-spaced checkpoint images on which to perform the"
          "analysis. Has no effect if images are also specified"
images = None
  .type = ints
  .help = "Images on which to perform the analysis (otherwise use all images)"
corrected = False
  .type = bool
  .help = "Use corrected data (i.e after applying pedestal and gain) in analysis"

masking {
  include scope dials.util.masking.phil_scope
}

output {
    plot = None
      .type = path
      .help = "Save background plot to file"
    size_inches = None
      .type = floats(value_min=0, size=2)
    log = dials.background.log
      .type = str
}

""",
    process_includes=True,
)


@show_mail_handle_errors()
def run(args=None):
    usage = "dials.background [options] image_*.cbf"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)

    # Ensure we have either a data block or an experiment list
    experiments = flatten_experiments(params.input.experiments)
    if not experiments:
        raise Sorry("No experiments found in the input")
    imagesets = experiments.imagesets()

    # Configure the logging
    log.config(logfile=params.output.log)
    logger.info(dials_version())

    if params.output.plot:
        import matplotlib

        matplotlib.use("agg")

        from matplotlib import pyplot

        from dials.util.matplotlib_utils import resolution_formatter

        fig = pyplot.figure(figsize=params.output.size_inches)
        ax = fig.add_subplot(111)

    for i_imgset, imageset in enumerate(imagesets):
        first, last = imageset.get_scan().get_image_range()
        images = range(first, last + 1)

        if params.n_checkpoints:
            if params.n_checkpoints == 1:
                images = (first,)
            elif params.n_checkpoints == 2:
                images = (first, last)
            else:
                step = max(1, int((last - first) / (params.n_checkpoints - 1)))
                images = range(first, last + 1, step)

        if params.images:
            if min(params.images) < first or max(params.images) > last:
                raise Sorry("image outside of scan range")
            images = params.images

        d_spacings = []
        intensities = []
        sigmas = []

        for indx in images:
            logger.info(f"For imageset {i_imgset} image {indx}:")
            d, I, sig = background(
                imageset,
                indx - first,  # indices passed to imageset.get_raw_data start from zero
                n_bins=params.n_bins,
                corrected=params.corrected,
                mask_params=params.masking,
                show_summary=True,
            )

            msg = [f"{'d':>8} {'I':>8} {'sig':>8}"]
            for j in range(len(I)):
                msg.append(f"{d[j]:8.3f} {I[j]:8.3f} {sig[j]:8.3f}")
            logger.info("\n".join(msg))

            d_spacings.append(d)
            intensities.append(I)
            sigmas.append(sig)

        if params.output.plot:
            ax.set_xlabel(r"resolution ($\AA$)")
            ax.set_ylabel(r"$\langle I_b \rangle$")
            for indx, d, I, sig in zip(images, d_spacings, intensities, sigmas):
                filenames = imageset.reader().paths()
                if len(imagesets) > 1:
                    label = (
                        f"{filenames[indx - first]}"
                        if len(filenames) > 1
                        else f"{filenames[0]} image {indx}"
                    )
                else:
                    label = f"image {indx}" if len(images) > 1 else ""
                ds2 = 1 / flex.pow2(d)
                ax.plot(ds2, I, label=label)
            ax.xaxis.set_major_formatter(resolution_formatter)

    if params.output.plot:
        try:
            if len(imagesets) > 1 or len(images) > 1:
                # Plot a legend if there are fewer lines than the number of colours
                # in the colour cycle
                if len(ax.lines) <= len(
                    pyplot.rcParams["axes.prop_cycle"].by_key()["color"]
                ):
                    pyplot.gca().legend()
            pyplot.savefig(params.output.plot)
        except ValueError:
            raise Sorry(f"Unable to save plot to {params.output.plot}")


def background(
    imageset, indx, n_bins, corrected=False, mask_params=None, show_summary=False
):
    if mask_params is None:
        # Default mask params for trusted range
        mask_params = phil_scope.fetch(parse("")).extract().masking

    detector = imageset.get_detector()
    beam = imageset.get_beam()

    # Only working with single panel detector for now
    assert len(detector) == 1
    panel = detector[0]
    imageset_mask = imageset.get_mask(indx)[0]
    mask = dials.util.masking.generate_mask(imageset, mask_params)[0]
    mask = imageset_mask & mask

    n = matrix.col(panel.get_normal()).normalize()
    b = matrix.col(beam.get_s0()).normalize()
    wavelength = beam.get_wavelength()

    if math.fabs(b.dot(n)) < 0.95:
        raise Sorry("Detector not perpendicular to beam")

    # Use corrected data to determine signal and background regions
    corrected_data = imageset.get_corrected_data(indx)
    assert len(corrected_data) == 1
    corrected_data = corrected_data[0].as_double()

    # Use choice of raw or corrected data to evaluate the background values
    if corrected:
        data = corrected_data
    else:
        data = imageset.get_raw_data(indx)[0].as_double()

    spot_params = spot_phil.fetch(source=parse("")).extract()
    threshold_function = SpotFinderFactory.configure_threshold(spot_params)
    peak_pixels = threshold_function.compute_threshold(corrected_data, mask)
    signal = data.select(peak_pixels.iselection())
    background_pixels = mask & ~peak_pixels
    background = data.select(background_pixels.iselection())

    # print some summary information
    if show_summary:
        logger.info(f"Mean background: {flex.sum(background) / background.size():.3f}")
        if len(signal) > 0:
            logger.info(
                f"Max/total signal pixels: {flex.max(signal):.0f} / {flex.sum(signal):.0f}"
            )
        else:
            logger.info("No signal pixels on this image")
        logger.info(
            "Peak/background/masked pixels: %d / %d / %d"
            % (peak_pixels.count(True), background.size(), mask.count(False))
        )

    # compute histogram of two-theta values, then same weighted
    # by pixel values, finally divide latter by former to get
    # the radial profile out, need to set the number of bins
    # sensibly; inspired by method in PyFAI

    two_theta_array = panel.get_two_theta_array(beam.get_s0())
    two_theta_array = two_theta_array.as_1d().select(background_pixels.iselection())

    # Use flex.weighted_histogram
    h0 = flex.weighted_histogram(two_theta_array, n_slots=n_bins)
    h1 = flex.weighted_histogram(two_theta_array, background, n_slots=n_bins)
    h2 = flex.weighted_histogram(
        two_theta_array, background * background, n_slots=n_bins
    )

    d0 = h0.slots()
    d1 = h1.slots()
    d2 = h2.slots()

    I = d1 / d0
    I2 = d2 / d0
    sig = flex.sqrt(I2 - flex.pow2(I))

    tt = h0.slot_centers()
    d_spacings = wavelength / (2.0 * flex.sin(0.5 * tt))

    return d_spacings, I, sig


if __name__ == "__main__":
    run()



from __future__ import annotations

import procrunner
import pytest

from iotbx import ccp4_map
from scitbx.array_family import flex


def test_rs_mapper(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.rs_mapper",
            dials_data("centroid_test_data", pathlib=True)
            / "imported_experiments.json",
            'map_file="junk.ccp4"',
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "junk.ccp4").is_file()

    # load results
    m = ccp4_map.map_reader(file_name=str(tmp_path / "junk.ccp4"))
    assert len(m.data) == 7189057
    assert m.header_min == 0.0
    assert flex.min(m.data) == 0.0

    assert m.header_max == 2052.75
    assert flex.max(m.data) == 2052.75

    assert m.header_mean == pytest.approx(0.018924040719866753, abs=1e-6)
    assert flex.mean(m.data) == pytest.approx(0.01892407052218914, abs=1e-6)


def test_masked(dials_data, tmp_path):
    procrunner.run(
        [
            "dials.import",
            dials_data("image_examples", pathlib=True) / "dectris_eiger_master.h5",
        ],
        working_directory=tmp_path,
    )
    result = procrunner.run(
        ["dials.rs_mapper", "imported.expt", "map_file=masked.ccp4"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "masked.ccp4").is_file()

    # Also check the ignore case (regressions here may indicate changes in mask handling)
    result = procrunner.run(
        [
            "dials.rs_mapper",
            "imported.expt",
            "map_file=unmasked.ccp4",
            "ignore_mask=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "unmasked.ccp4").is_file()

    # Load results
    masked = ccp4_map.map_reader(file_name=str(tmp_path / "masked.ccp4"))
    unmasked = ccp4_map.map_reader(file_name=str(tmp_path / "unmasked.ccp4"))

    assert masked.header_max < unmasked.header_max
    assert masked.header_max == pytest.approx(289.11111)
    assert unmasked.header_max == pytest.approx(65535.0)



"""
Test refinement of a crystal unit cell using a two theta target.
"""


from __future__ import annotations

import os
from copy import deepcopy
from math import pi

from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.test_utils import approx_equal

from dials.algorithms.refinement.two_theta_refiner import (
    TwoThetaExperimentsPredictor,
    TwoThetaPredictionParameterisation,
    TwoThetaReflectionManager,
    TwoThetaTarget,
)


def generate_reflections(experiments):

    from cctbx.sgtbx import space_group, space_group_symbols
    from scitbx.array_family import flex

    from dials.algorithms.refinement.prediction.managed_predictors import (
        ScansExperimentsPredictor,
        ScansRayPredictor,
    )
    from dials.algorithms.spot_prediction import IndexGenerator, ray_intersection

    detector = experiments[0].detector
    crystal = experiments[0].crystal

    # All indices in a 2.0 Angstrom sphere
    resolution = 2.0
    index_generator = IndexGenerator(
        crystal.get_unit_cell(),
        space_group(space_group_symbols(1).hall()).type(),
        resolution,
    )
    indices = index_generator.to_array()

    # Predict rays within the sequence range
    scan = experiments[0].scan
    sequence_range = scan.get_oscillation_range(deg=False)
    ray_predictor = ScansRayPredictor(experiments, sequence_range)
    obs_refs = ray_predictor(indices)

    # Take only those rays that intersect the detector
    intersects = ray_intersection(detector, obs_refs)
    obs_refs = obs_refs.select(intersects)

    # Make a reflection predictor and re-predict for all these reflections. The
    # result is the same, but we gain also the flags and xyzcal.px columns
    ref_predictor = ScansExperimentsPredictor(experiments)
    obs_refs["id"] = flex.int(len(obs_refs), 0)
    obs_refs = ref_predictor(obs_refs)

    # Set 'observed' centroids from the predicted ones
    obs_refs["xyzobs.mm.value"] = obs_refs["xyzcal.mm"]

    # Invent some variances for the centroid positions of the simulated data
    im_width = 0.1 * pi / 180.0
    px_size = detector[0].get_pixel_size()
    var_x = flex.double(len(obs_refs), (px_size[0] / 2.0) ** 2)
    var_y = flex.double(len(obs_refs), (px_size[1] / 2.0) ** 2)
    var_phi = flex.double(len(obs_refs), (im_width / 2.0) ** 2)
    obs_refs["xyzobs.mm.variance"] = flex.vec3_double(var_x, var_y, var_phi)

    return obs_refs, ref_predictor


def test_fd_derivatives():
    """Test derivatives of the prediction equation"""

    from libtbx.phil import parse

    from . import geometry_phil

    # Import model builder
    from .setup_geometry import Extract

    # Create models
    overrides = """geometry.parameters.crystal.a.length.range = 10 50
  geometry.parameters.crystal.b.length.range = 10 50
  geometry.parameters.crystal.c.length.range = 10 50"""
    master_phil = parse(geometry_phil)
    models = Extract(master_phil, overrides)

    mydetector = models.detector
    mygonio = models.goniometer
    mycrystal = models.crystal
    mybeam = models.beam

    # Build a mock scan for a 72 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    myscan = sf.make_scan(
        image_range=(1, 720),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(720)),
        deg=True,
    )

    # Create a parameterisation of the crystal unit cell
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalUnitCellParameterisation,
    )

    xluc_param = CrystalUnitCellParameterisation(mycrystal)

    # Create an ExperimentList
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=mybeam,
            detector=mydetector,
            goniometer=mygonio,
            scan=myscan,
            crystal=mycrystal,
            imageset=None,
        )
    )

    # Build a prediction parameterisation for two theta prediction
    pred_param = TwoThetaPredictionParameterisation(
        experiments,
        detector_parameterisations=None,
        beam_parameterisations=None,
        xl_orientation_parameterisations=None,
        xl_unit_cell_parameterisations=[xluc_param],
    )

    # Generate some reflections
    obs_refs, ref_predictor = generate_reflections(experiments)

    # Build a ReflectionManager with overloads for handling 2theta residuals
    refman = TwoThetaReflectionManager(obs_refs, experiments, outlier_detector=None)

    # Build a TwoThetaExperimentsPredictor
    ref_predictor = TwoThetaExperimentsPredictor(experiments)

    # Make a target for the least squares 2theta residual
    target = TwoThetaTarget(experiments, ref_predictor, refman, pred_param)

    # Keep only reflections that pass inclusion criteria and have predictions
    reflections = refman.get_matches()

    # Get analytical gradients
    an_grads = pred_param.get_gradients(reflections)

    # Get finite difference gradients
    p_vals = pred_param.get_param_vals()
    deltas = [1.0e-7] * len(p_vals)

    for i in range(len(deltas)):

        val = p_vals[i]

        p_vals[i] -= deltas[i] / 2.0
        pred_param.set_param_vals(p_vals)

        target.predict()
        reflections = refman.get_matches()

        rev_state = reflections["2theta_resid"].deep_copy()

        p_vals[i] += deltas[i]
        pred_param.set_param_vals(p_vals)

        target.predict()
        reflections = refman.get_matches()

        fwd_state = reflections["2theta_resid"].deep_copy()
        p_vals[i] = val

        fd = fwd_state - rev_state
        fd /= deltas[i]

        # compare with analytical calculation
        assert approx_equal(fd, an_grads[i]["d2theta_dp"], eps=1.0e-6)

    # return to the initial state
    pred_param.set_param_vals(p_vals)


def test_refinement(dials_regression):
    """Test a refinement run"""

    # Get a beam and detector from a experiments. This one has a CS-PAD, but that
    # is irrelevant
    data_dir = os.path.join(dials_regression, "refinement_test_data", "hierarchy_test")
    experiments_path = os.path.join(data_dir, "datablock.json")
    assert os.path.exists(experiments_path)

    # load models
    from dxtbx.model.experiment_list import ExperimentListFactory

    experiments = ExperimentListFactory.from_serialized_format(
        experiments_path, check_format=False
    )
    im_set = experiments.imagesets()[0]
    detector = deepcopy(im_set.get_detector())
    beam = im_set.get_beam()

    # Invent a crystal, goniometer and scan for this test
    from dxtbx.model import Crystal

    crystal = Crystal(
        (40.0, 0.0, 0.0), (0.0, 40.0, 0.0), (0.0, 0.0, 40.0), space_group_symbol="P1"
    )
    orig_xl = deepcopy(crystal)

    from dxtbx.model import GoniometerFactory

    goniometer = GoniometerFactory.known_axis((1.0, 0.0, 0.0))

    # Build a mock scan for a 180 degree sequence
    from dxtbx.model import ScanFactory

    sf = ScanFactory()
    scan = sf.make_scan(
        image_range=(1, 1800),
        exposure_times=0.1,
        oscillation=(0, 0.1),
        epochs=list(range(1800)),
        deg=True,
    )
    sequence_range = scan.get_oscillation_range(deg=False)
    im_width = scan.get_oscillation(deg=False)[1]
    assert sequence_range == (0.0, pi)
    assert approx_equal(im_width, 0.1 * pi / 180.0)

    # Build an experiment list
    experiments = ExperimentList()
    experiments.append(
        Experiment(
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            crystal=crystal,
            imageset=None,
        )
    )

    # simulate some reflections
    refs, _ = generate_reflections(experiments)

    # change unit cell a bit (=0.1 Angstrom length upsets, 0.1 degree of
    # alpha and beta angles)
    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalUnitCellParameterisation,
    )

    xluc_param = CrystalUnitCellParameterisation(crystal)
    cell_params = crystal.get_unit_cell().parameters()
    cell_params = [a + b for a, b in zip(cell_params, [0.1, -0.1, 0.1, 0.1, -0.1, 0.0])]
    from cctbx.uctbx import unit_cell
    from rstbx.symmetry.constraints.parameter_reduction import symmetrize_reduce_enlarge
    from scitbx import matrix

    new_uc = unit_cell(cell_params)
    newB = matrix.sqr(new_uc.fractionalization_matrix()).transpose()
    S = symmetrize_reduce_enlarge(crystal.get_space_group())
    S.set_orientation(orientation=newB)
    X = tuple([e * 1.0e5 for e in S.forward_independent_parameters()])
    xluc_param.set_param_vals(X)

    # reparameterise the crystal at the perturbed geometry
    xluc_param = CrystalUnitCellParameterisation(crystal)

    # Dummy parameterisations for other models
    beam_param = None
    xlo_param = None
    det_param = None

    # parameterisation of the prediction equation
    from dials.algorithms.refinement.parameterisation.parameter_report import (
        ParameterReporter,
    )

    pred_param = TwoThetaPredictionParameterisation(
        experiments, det_param, beam_param, xlo_param, [xluc_param]
    )
    param_reporter = ParameterReporter(det_param, beam_param, xlo_param, [xluc_param])

    # reflection manager
    refman = TwoThetaReflectionManager(refs, experiments, nref_per_degree=20)

    # reflection predictor
    ref_predictor = TwoThetaExperimentsPredictor(experiments)

    # target function
    target = TwoThetaTarget(experiments, ref_predictor, refman, pred_param)

    # minimisation engine
    from dials.algorithms.refinement.engine import (
        LevenbergMarquardtIterations as Refinery,
    )

    refinery = Refinery(
        target=target,
        prediction_parameterisation=pred_param,
        log=None,
        max_iterations=20,
    )

    # Refiner
    from dials.algorithms.refinement.refiner import Refiner

    refiner = Refiner(
        experiments=experiments,
        pred_param=pred_param,
        param_reporter=param_reporter,
        refman=refman,
        target=target,
        refinery=refinery,
    )
    refiner.run()

    # compare crystal with original crystal
    refined_xl = refiner.get_experiments()[0].crystal

    # print refined_xl
    assert refined_xl.is_similar_to(
        orig_xl, uc_rel_length_tolerance=0.001, uc_abs_angle_tolerance=0.01
    )

    # print "Unit cell esds:"
    # print refined_xl.get_cell_parameter_sd()



from __future__ import annotations


def enhash(e, h, k, l):
    return e * (2**30) + (h + 512) * (2**20) + (k + 512) * (2**10) + (l + 512)


def add_hash(integrated_data):
    """Add hash = 2^30 * entering + 2^20 * (h+512) + 2^10 * (k+512) + (l+512)
    as new column to reflection table - should be P1 unique for 360 degree
    scans"""

    integrated_data = integrated_data.select(integrated_data["id"] >= 0)
    assert max(integrated_data["id"]) == 0

    h, k, l = integrated_data["miller_index"].as_vec3_double().parts()
    h = h.iround()
    k = k.iround()
    l = l.iround()
    e = integrated_data["entering"].as_int()

    hash_value = enhash(e, h, k, l)
    integrated_data["hash"] = hash_value

    return integrated_data


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        raise RuntimeError(f"{sys.argv[0]} strong.refl")

    from dials.array_family import flex

    integrated_data = flex.reflection_table.from_file(sys.argv[1])

    # keep only flagged as integrated reflections
    sel = integrated_data.get_flags(integrated_data.flags.integrated)
    integrated_data = integrated_data.select(sel)

    # only fully recorded reflections
    sel = integrated_data["partiality"] > 0.99
    integrated_data = integrated_data.select(sel)

    integrated_data = add_hash(integrated_data)
    hash_value = integrated_data["hash"]

    for h in hash_value:
        sel = hash_value == h
        assert sel.count(True) == 1

    print(flex.min(hash_value), flex.max(hash_value))



from __future__ import annotations

from libtbx.phil import parse

phil_scope = parse(
    """

mp {
  method = *multiprocessing sge lsf pbs
    .type = choice
    .help = "The multiprocessing method to use"

  nproc = 1
    .type = int(value_min=1)
    .help = "The number of processes to use."

  nthreads = 1
    .type = int(value_min=1)
    .help = "The number of local threads to use for openmp."
}
"""
)



from __future__ import annotations

import math

import pytest

from dxtbx.model import Goniometer
from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.array_family import flex

from dials.algorithms.refinement.parameterisation.goniometer_parameters import (
    GoniometerParameterisation,
)
from dials.algorithms.refinement.refinement_helpers import (
    get_fd_gradients,
    random_param_shift,
)


def random_gonio():
    # make a random rotation axis with a random setting matrix
    axis = matrix.col(flex.random_double_point_on_sphere())
    fixed_rotation = matrix.sqr((1, 0, 0, 0, 1, 0, 0, 0, 1))
    setting_rotation = matrix.sqr(flex.random_double_r3_rotation_matrix())
    goniometer = Goniometer(axis, fixed_rotation, setting_rotation)
    return goniometer


def test():
    goniometer = random_gonio()
    gonp = GoniometerParameterisation(goniometer)

    # Let's do some basic tests. First, can we change parameter values and
    # update the laboratory frame rotation axis?
    e_lab = matrix.col(goniometer.get_rotation_axis())
    gonp.set_param_vals([1000 * 0.1, 1000 * 0.1])
    assert matrix.col(goniometer.get_rotation_axis()).angle(e_lab) == pytest.approx(
        0.1413033
    )

    # random goniometers and random parameter shifts
    attempts = 1000
    for i in range(attempts):
        # make a random goniometer and parameterise it
        goniometer = random_gonio()
        gonp = GoniometerParameterisation(goniometer)

        # apply a random parameter shift
        p_vals = gonp.get_param_vals()
        p_vals = random_param_shift(p_vals, [1000 * math.pi / 9, 1000 * math.pi / 9])
        gonp.set_param_vals(p_vals)

        # compare analytical and finite difference derivatives
        an_ds_dp = gonp.get_ds_dp()
        fd_ds_dp = get_fd_gradients(
            gonp, [1.0e-5 * math.pi / 180, 1.0e-5 * math.pi / 180]
        )

        null_mat = matrix.sqr((0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0))
        for j in range(2):
            try:
                assert approx_equal((fd_ds_dp[j] - an_ds_dp[j]), null_mat, eps=1.0e-6)
            except Exception:
                print("for try", i)
                print("failure for parameter number", j)
                print("with fd_ds_dp = ")
                print(fd_ds_dp[j])
                print("and an_ds_dp = ")
                print(an_ds_dp[j])
                print("so that difference fd_ds_dp - an_ds_dp =")
                print(fd_ds_dp[j] - an_ds_dp[j])
                raise



from __future__ import annotations

import math
import pickle
import random


def test_getters():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    image_size = sampler.image_size()
    image_centre = sampler.image_centre()
    r0 = sampler.r0()
    r1 = sampler.r1()
    r2 = sampler.r2()
    size = len(sampler)

    assert width == image_size[0]
    assert height == image_size[1]
    assert width // 2 == image_centre[0]
    assert height // 2 == image_centre[1]
    assert r0 == min([width // 2, height // 2])
    assert r1 == r0 / 3.0
    assert r2 == r1 * math.sqrt(5.0)
    assert 9 * nz == size


def test_detector_area():
    from scitbx.array_family import flex

    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    im = flex.int(flex.grid(height, width))
    for j in range(height):
        for i in range(width):
            im[j, i] = sampler.nearest(0, (i, j, 0))

    assert im[height // 2, width // 2] == 0
    assert im[height // 2, width - 1] == 1
    assert im[height - 1, width - 1] == 2
    assert im[height - 1, width // 2] == 3
    assert im[height - 1, 0] == 4
    assert im[height // 2, 0] == 5
    assert im[0, 0] == 6
    assert im[0, width // 2] == 7
    assert im[0, width - 1] == 8


#        from matplotlib import pylab
#        pylab.imshow(im.as_numpy_array())
#        pylab.show()


def test_indexing():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    zstep = depth / nz
    eps = 1e-10

    dt = 2 * math.pi / 8
    r2 = sampler.r2()

    xp = [width / 2.0] + [width / 2.0 + r2 * math.cos(i * dt) for i in range(8)]
    yp = [height / 2.0] + [height / 2.0 + r2 * math.sin(i * dt) for i in range(8)]
    zind = [[k] * 9 for k in range(nz)]
    zind = [i for j in zind for i in j]
    zp = [(z + 0.5) * zstep + scan_range[0] for z in zind]

    for x0, y0, z0, i in zip(xp, yp, zp, range(len(sampler))):
        x1, y1, z1 = sampler.coord(i)
        assert abs(x0 - x1) <= eps
        assert abs(y0 - y1) <= eps
        assert abs(z0 - z1) <= eps


def test_nearest():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    xc, yc = sampler.image_centre()
    r1 = sampler.r1()

    for i in range(1000):
        x = random.uniform(0, 1000)
        y = random.uniform(0, 1000)
        z = random.uniform(scan_range[0], scan_range[1])

        r = math.sqrt((x - xc) ** 2 + (y - yc) ** 2)
        if r < r1:
            index00 = 0
        else:
            t = math.atan2(y - yc, x - xc)
            ai = int(math.floor(t * 8 / (2 * math.pi) + 0.5)) % 8
            index00 = ai + 1

        index01 = int((z - scan_range[0]) * nz / depth)
        if index01 < 0:
            index01 = 0
        if index01 >= 2:
            index01 = 1

        index0 = index00 + index01 * 9
        index1 = sampler.nearest(0, (x, y, z))
        assert index0 == index1


def test_nearest_n():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    depth = scan_range[1] - scan_range[0]
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    xc, yc = sampler.image_centre()
    r1 = sampler.r1()

    for i in range(1000):
        x = random.uniform(0, 1000)
        y = random.uniform(0, 1000)
        z = random.uniform(scan_range[0], scan_range[1])

        r = math.sqrt((x - xc) ** 2 + (y - yc) ** 2)
        if r < r1:
            index00 = 0
        else:
            t = math.atan2(y - yc, x - xc)
            ai = int(math.floor(t * 8 / (2 * math.pi) + 0.5)) % 8
            index00 = ai + 1

        index01 = int((z - scan_range[0]) * nz / depth)
        if index01 < 0:
            index01 = 0
        if index01 >= 2:
            index01 = 1

        index0 = index00 + index01 * 9
        index1 = sampler.nearest_n(0, (x, y, z))
        assert index0 == index1[0]
        if index0 % 9 == 0:
            assert len(index1) == 9
            assert all(idx == index0 + i for i, idx in enumerate(index1))
        else:
            assert len(index1) == 4
            assert index1[1] == (index0 // 9) * 9
            if (index0 % 9) == 1:
                assert index1[2] == index0 + 1
                assert index1[3] == index0 + 7
            elif (index0 % 9) == 8:
                assert index1[2] == index0 - 7
                assert index1[3] == index0 - 1
            else:
                assert index1[2] == index0 + 1
                assert index1[3] == index0 - 1


def test_weights():
    from scitbx import matrix

    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 1
    sampler = CircleSampler((width, height), scan_range, nz)

    # Check the weight at the coord in 1.0
    eps = 1e-7
    for i in range(len(sampler)):
        coord = sampler.coord(i)
        weight = sampler.weight(i, 0, coord)
        assert abs(weight - 1.0) < eps

    r1 = sampler.r1()
    r2 = sampler.r2()
    r = r2 / (2.0 * r1)
    expected = math.exp(-4.0 * r * r * math.log(2.0))
    for i in range(1, 9):
        coord = sampler.coord(i)
        weight = sampler.weight(0, 0, coord)
        assert abs(weight - expected) < eps

    r = r2 / (2.0 * (r2 - r1))
    expected = math.exp(-4.0 * r * r * math.log(2.0))
    for i in range(1, 9):
        coord = sampler.coord(0)
        weight = sampler.weight(i, 0, coord)
        assert abs(weight - expected) < eps

    for i in range(1, 9):
        coord1 = matrix.col(sampler.coord(0))
        coord2 = matrix.col(sampler.coord(i))
        coord = coord1 + r1 * (coord2 - coord1) / r2
        weight = sampler.weight(0, 0, coord)
        assert abs(weight - 0.5) < eps
        weight = sampler.weight(i, 0, coord)
        assert abs(weight - 0.5) < eps


def test_self_consistent():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)

    for i in range(len(sampler)):
        coord = sampler.coord(i)
        index = sampler.nearest(0, coord)
        assert index == i


def test_z_index():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)
    assert (sampler.nearest(0, (500, 500, 2.0)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 3.0)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 4.0)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 5.0)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 6.0)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 6.5)) / 9) == 0
    assert (sampler.nearest(0, (500, 500, 7.0)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 7.5)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 8.0)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 9.0)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 10.0)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 11.0)) / 9) == 1
    assert (sampler.nearest(0, (500, 500, 12.0)) / 9) == 1


def test_pickle():
    from dials.algorithms.profile_model.modeller import CircleSampler

    width = 1000
    height = 1000
    scan_range = (2, 12)
    nz = 2
    sampler = CircleSampler((width, height), scan_range, nz)

    sampler2 = pickle.loads(pickle.dumps(sampler))

    assert sampler.image_size() == sampler2.image_size()
    assert sampler.num_z() == sampler2.num_z()



from __future__ import annotations

import glob
import os

import pytest

import scitbx
from cctbx import uctbx
from dxtbx.model import ExperimentList
from dxtbx.serialize import load

from dials.command_line import search_beam_position

from ..algorithms.indexing.test_index import run_indexing


def test_search_i04_weak_data_image_range(mocker, run_in_tmp_path, dials_regression):
    """Perform a beam-centre search and check that the output is sane."""

    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    reflection_file = os.path.join(data_dir, "full.pickle")
    experiments_file = os.path.join(data_dir, "experiments_import.json")

    args = [
        experiments_file,
        reflection_file,
        "image_range=1,10",
        "image_range=251,260",
        "image_range=531,540",
        "n_macro_cycles=4",
    ]
    from rstbx.indexing_api import dps_extended

    mocker.spy(dps_extended, "get_new_detector")
    search_beam_position.run(args)
    # Check that the last call to get_new_detector was with an offset of close to zero.
    # The final call was to apply the "best" shift to the detector model before
    # returning the updated experiments.
    assert dps_extended.get_new_detector.call_args[0][1].elems == pytest.approx(
        (0, 0, 0), abs=3e-2
    )
    assert os.path.exists("optimised.expt")

    # Compare the shifts between the start and final detector models
    experiments = load.experiment_list(experiments_file, check_format=False)
    optimised_experiments = load.experiment_list("optimised.expt", check_format=False)
    detector_1 = experiments[0].detector
    detector_2 = optimised_experiments[0].detector
    shift = scitbx.matrix.col(detector_1[0].get_origin()) - scitbx.matrix.col(
        detector_2[0].get_origin()
    )
    assert shift.elems == pytest.approx((0.27, -0.12, 0.0), abs=1e-1)


def test_search_multiple(run_in_tmp_path, dials_regression):
    """Perform a beam-centre search and check that the output is sane.

    Do the following:
    1. Run dials.search_beam_centre on two datablocks and two pickled
    reflection tables, as output by dials.find_spots;
      a) Check that the program exits correctly;
      b) Check that it produces the expected output datablock.
    2. Check that the beam centre search has resulted in the expected shift
    in detector origin.
    """

    data_dir = os.path.join(dials_regression, "indexing_test_data", "trypsin")
    pickle_path1 = os.path.join(data_dir, "strong_P1_X6_1_0-1.pickle")
    pickle_path2 = os.path.join(data_dir, "strong_P1_X6_2_0-1.pickle")
    experiments_path1 = os.path.join(data_dir, "datablock_P1_X6_1.json")
    experiments_path2 = os.path.join(data_dir, "datablock_P1_X6_2.json")

    args = [experiments_path1, experiments_path2, pickle_path1, pickle_path2]
    search_beam_position.run(args)
    assert os.path.exists("optimised.expt")

    experiments = load.experiment_list(experiments_path1, check_format=False)
    optimised_experiments = load.experiment_list("optimised.expt", check_format=False)
    detector_1 = experiments[0].detector
    detector_2 = optimised_experiments[0].detector
    shift = scitbx.matrix.col(detector_1[0].get_origin()) - scitbx.matrix.col(
        detector_2[0].get_origin()
    )
    assert shift.elems == pytest.approx((-0.518, 0.192, 0.0), abs=1e-1)


def test_index_after_search(dials_data, run_in_tmp_path):
    """Integrate the beam centre search with the rest of the toolchain

    Do the following:
    1. Take a known good experiment and perturbate the beam centre
    2. Run dials.search_beam_centre on the perturbated beam centre and original
    reflection table, check for expected output;
    3. Run dials.index with the found beam centre and check that the expected
    unit cell is obtained and that the RMSDs are smaller than or equal to some
    expected values."""

    insulin = dials_data("insulin_processed", pathlib=True)

    # load the original experiment and perturbate the beam centre by a small offset
    experiments = load.experiment_list(insulin / "imported.expt", check_format=False)
    original_origin = experiments[0].detector.hierarchy().get_origin()
    shifted_origin = (
        original_origin[0] - 1.3,
        original_origin[1] + 1.5,
        original_origin[2],
    )
    experiments[0].detector.hierarchy().set_local_frame(
        experiments[0].detector.hierarchy().get_fast_axis(),
        experiments[0].detector.hierarchy().get_slow_axis(),
        shifted_origin,
    )
    assert experiments[0].detector.hierarchy().get_origin() == shifted_origin
    experiments.as_file(run_in_tmp_path / "shifted.expt")

    # search the beam centre
    search_beam_position.run(
        [
            str(run_in_tmp_path / "shifted.expt"),
            str(insulin / "strong.refl"),
        ]
    )
    assert run_in_tmp_path.joinpath("optimised.expt").is_file()

    # check we can actually index the resulting optimized experiments
    expected_unit_cell = uctbx.unit_cell(
        (67.655, 67.622, 67.631, 109.4583, 109.4797, 109.485)
    )
    expected_rmsds = (0.3, 0.3, 0.005)
    expected_hall_symbol = " P 1"
    run_indexing(
        insulin / "strong.refl",
        run_in_tmp_path / "optimised.expt",
        run_in_tmp_path,
        [],
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_search_single(run_in_tmp_path, dials_regression):
    """Perform a beam-centre search and check that the output is sane.

    Do the following:
    1. Run dials.search_beam_centre on a single datablock and pickled
    reflection table, as output by dials.find_spots;
      a) Check that the program exits correctly;
      b) Check that it produces the expected output datablock.
    2. Check that the beam centre search has resulted in the expected shift
    in detector origin.
    """

    data_dir = os.path.join(dials_regression, "indexing_test_data", "phi_scan")
    pickle_path = os.path.join(data_dir, "strong.pickle")
    experiments_path = os.path.join(data_dir, "datablock.json")

    search_beam_position.run([experiments_path, pickle_path])
    assert run_in_tmp_path.joinpath("optimised.expt").is_file()

    experiments = load.experiment_list(experiments_path, check_format=False)
    original_imageset = experiments.imagesets()[0]
    optimized_experiments = load.experiment_list("optimised.expt", check_format=False)
    detector_1 = original_imageset.get_detector()
    detector_2 = optimized_experiments.detectors()[0]
    shift = scitbx.matrix.col(detector_1[0].get_origin()) - scitbx.matrix.col(
        detector_2[0].get_origin()
    )
    assert shift.elems == pytest.approx((-0.976, 2.497, 0.0), abs=1e-1)


def test_search_small_molecule(dials_data, run_in_tmp_path):
    """Perform a beam-centre search on a multi-sequence data set..

    Do the following:
    1. Run dials.search_beam_centre on a single datablock and pickled
    reflection table containing multiple experiment IDs, as output by
    dials.find_spots;
      a) Check that the program exits correctly;
      b) Check that it produces the expected output datablock.
    2. Check that the beam centre search has resulted in the expected shift
    in detector origin.
    """

    data = dials_data("l_cysteine_dials_output", pathlib=True)
    experiments_path = data / "imported.expt"
    refl_path = data / "strong.refl"

    search_beam_position.run([os.fspath(experiments_path), os.fspath(refl_path)])
    assert run_in_tmp_path.joinpath("optimised.expt").is_file()

    experiments = load.experiment_list(experiments_path, check_format=False)
    optimised_experiments = load.experiment_list("optimised.expt", check_format=False)
    for old_expt, new_expt in zip(experiments, optimised_experiments):
        # assert that the detector fast/slow axes are unchanged from the input experiments
        # the last experiment actually does have a different detector model
        assert (
            old_expt.detector[0].get_slow_axis() == new_expt.detector[0].get_slow_axis()
        )
        assert (
            old_expt.detector[0].get_fast_axis() == new_expt.detector[0].get_fast_axis()
        )
        shift = scitbx.matrix.col(
            old_expt.detector[0].get_origin()
        ) - scitbx.matrix.col(new_expt.detector[0].get_origin())
        assert shift.elems == pytest.approx((0.091, -1.11, 0), abs=1e-2)


def test_multi_sweep_fixed_rotation(dials_regression, run_in_tmp_path):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "multi_sweep")
    reflection_files = sorted(
        glob.glob(os.path.join(data_dir, "SWEEP[1,2]", "index", "*_strong.pickle"))
    )
    experiment_files = sorted(
        glob.glob(
            os.path.join(data_dir, "SWEEP[1,2]", "index", "*_datablock_import.json")
        )
    )

    search_beam_position.run(reflection_files + experiment_files)
    assert run_in_tmp_path.joinpath("optimised.expt").is_file()

    experiments = ExperimentList()
    for path in experiment_files:
        experiments.extend(load.experiment_list(path, check_format=False))

    optimised_experiments = load.experiment_list("optimised.expt", check_format=False)
    for orig_expt, new_expt in zip(experiments, optimised_experiments):
        shift = scitbx.matrix.col(
            orig_expt.detector[0].get_origin()
        ) - scitbx.matrix.col(new_expt.detector[0].get_origin())
        print(shift)
        assert shift.elems == pytest.approx((2.293, -0.399, 0), abs=1e-2)



"""
Figure out correct gradient expressions required for crystal unit cell
restraints
"""


from __future__ import annotations

import math
import os


def test(dials_regression):
    from libtbx.phil import parse
    from libtbx.test_utils import approx_equal
    from scitbx import matrix

    from dials.algorithms.refinement.parameterisation.crystal_parameters import (
        CrystalOrientationParameterisation,
        CrystalUnitCellParameterisation,
    )

    # Get modules to build models and minimiser using PHIL
    from . import geometry_phil, minimiser_phil, setup_geometry

    # Symmetry constrained parameterisation for the unit cell
    DEG2RAD = math.pi / 180.0
    RAD2DEG = 180.0 / math.pi

    master_phil = parse(f"{geometry_phil}\n{minimiser_phil}")

    # make cell more oblique
    args = [
        "a.direction.close_to.sd=5",
        "b.direction.close_to.sd=5",
        "c.direction.close_to.sd=5",
    ]
    models = setup_geometry.Extract(master_phil, cmdline_args=args)
    crystal = models.crystal

    # a hexagonal crystal is a good test case for behaviour of oblique cells
    do_hexagonal = True
    if do_hexagonal:
        from dxtbx.model.experiment_list import ExperimentListFactory

        experiments = ExperimentListFactory.from_json_file(
            os.path.join(
                dials_regression,
                "refinement_test_data",
                "multi_stills",
                "combined_experiments.json",
            ),
            check_format=False,
        )
        crystal = experiments[0].crystal

    # derive finite difference gradients of various quantities wrt each param
    def check_fd_gradients(parameterisation):
        mp = parameterisation
        p_vals = mp.get_param_vals()
        deltas = [1.0e-7 for p in p_vals]
        assert len(deltas) == len(p_vals)
        fd_grad = []

        # get matrix to unset rotations of unit cell vectors
        Ut = matrix.sqr(mp.get_model().get_U()).transpose()

        for i, delta in enumerate(deltas):
            val = p_vals[i]

            p_vals[i] -= delta / 2.0
            mp.set_param_vals(p_vals)
            rev_uc = mp.get_model().get_unit_cell().parameters()
            rev_vec = mp.get_model().get_real_space_vectors()
            rev_vec = [Ut * vec for vec in rev_vec]
            rev_B = matrix.sqr(mp.get_model().get_B())
            rev_O = rev_B.transpose().inverse()

            p_vals[i] += delta
            mp.set_param_vals(p_vals)
            fwd_uc = mp.get_model().get_unit_cell().parameters()
            fwd_vec = mp.get_model().get_real_space_vectors()
            fwd_vec = [Ut * vec for vec in fwd_vec]
            fwd_B = matrix.sqr(mp.get_model().get_B())
            fwd_O = fwd_B.transpose().inverse()

            fd_uc = [(f - r) / delta for f, r in zip(fwd_uc, rev_uc)]
            fd_vec = [(f - r) / delta for f, r in zip(fwd_vec, rev_vec)]
            fd_B = (fwd_B - rev_B) / delta
            fd_O = (fwd_O - rev_O) / delta

            fd_grad.append(
                {
                    "da_dp": fd_uc[0],
                    "db_dp": fd_uc[1],
                    "dc_dp": fd_uc[2],
                    "daa_dp": fd_uc[3],
                    "dbb_dp": fd_uc[4],
                    "dcc_dp": fd_uc[5],
                    "davec_dp": fd_vec[0],
                    "dbvec_dp": fd_vec[1],
                    "dcvec_dp": fd_vec[2],
                    "dB_dp": fd_B,
                    "dO_dp": fd_O,
                }
            )

            p_vals[i] = val

        # return to the initial state
        mp.set_param_vals(p_vals)

        return fd_grad

    assert CrystalOrientationParameterisation(crystal)
    xluc_param = CrystalUnitCellParameterisation(crystal)

    from dials.algorithms.refinement.restraints.restraints import SingleUnitCellTie

    assert SingleUnitCellTie(xluc_param, [0] * 6, [0] * 6)

    from scitbx.math import angle_derivative_wrt_vectors

    B = matrix.sqr(crystal.get_B())
    O = (B.transpose()).inverse()
    a, b, c, aa, bb, cc = crystal.get_unit_cell().parameters()
    aa *= DEG2RAD
    bb *= DEG2RAD
    cc *= DEG2RAD
    Ut = matrix.sqr(crystal.get_U()).transpose()
    avec, bvec, cvec = [Ut * vec for vec in crystal.get_real_space_vectors()]

    # calculate d[B^T]/dp
    dB_dp = xluc_param.get_ds_dp()
    dBT_dp = [dB.transpose() for dB in dB_dp]

    # calculate d[O]/dp
    dO_dp = [-O * dBT * O for dBT in dBT_dp]

    # function to get analytical derivative of angles wrt vectors
    def dangle(u, v):
        return [matrix.col(e) for e in angle_derivative_wrt_vectors(u, v)]

    dalpha_db, dalpha_dc = dangle(bvec, cvec)
    dbeta_da, dbeta_dc = dangle(avec, cvec)
    dgamma_da, dgamma_db = dangle(avec, bvec)

    # get all FD derivatives
    fd_grad = check_fd_gradients(xluc_param)

    # look at each parameter
    for i, dO in enumerate(dO_dp):

        # print
        # print "***** PARAMETER {0} *****".format(i)

        # print "dB_dp analytical"
        # print dB_dp[i]
        # print "dB_dp FD"
        # print fd_grad[i]['dB_dp']
        # print

        # dB_dp is good. What about dO_dp?

        # print "O MATRIX"
        # print "dO_dp analytical"
        # print dO.round(6)
        # print "dO_dp FD"
        # print fd_grad[i]['dO_dp'].round(6)
        # print
        assert approx_equal(dO, fd_grad[i]["dO_dp"])

        # extract derivatives of each unit cell vector wrt p
        dav_dp, dbv_dp, dcv_dp = dO.transpose().as_list_of_lists()
        dav_dp = matrix.col(dav_dp)
        dbv_dp = matrix.col(dbv_dp)
        dcv_dp = matrix.col(dcv_dp)

        # check these are correct vs FD
        # print "CELL VECTORS"
        # diff = dav_dp - fd_grad[i]['davec_dp']
        # print 2 * diff.length() / (dav_dp.length() + fd_grad[i]['davec_dp'].length()) * 100
        # print 'davec_dp analytical: {0:.5f} {1:.5f} {2:.5f}'.format(*dav_dp.elems)
        # print 'davec_dp finite diff: {0:.5f} {1:.5f} {2:.5f}'.format(*fd_grad[i]['davec_dp'].elems)
        assert approx_equal(dav_dp, fd_grad[i]["davec_dp"])

        # diff = dbv_dp - fd_grad[i]['dbvec_dp']
        # print 2 * diff.length() / (dbv_dp.length() + fd_grad[i]['dbvec_dp'].length()) * 100
        # print 'dbvec_dp analytical: {0:.5f} {1:.5f} {2:.5f}'.format(*dbv_dp.elems)
        # print 'dbvec_dp finite diff: {0:.5f} {1:.5f} {2:.5f}'.format(*fd_grad[i]['dbvec_dp'].elems)
        assert approx_equal(dbv_dp, fd_grad[i]["dbvec_dp"])

        # diff = dcv_dp - fd_grad[i]['dcvec_dp']
        # print 2 * diff.length() / (dcv_dp.length() + fd_grad[i]['dcvec_dp'].length()) * 100
        # print 'dcvec_dp analytical: {0:.5f} {1:.5f} {2:.5f}'.format(*dcv_dp.elems)
        # print 'dcvec_dp finite diff: {0:.5f} {1:.5f} {2:.5f}'.format(*fd_grad[i]['dcvec_dp'].elems)
        # print
        assert approx_equal(dcv_dp, fd_grad[i]["dcvec_dp"])

        # print "CELL LENGTHS"
        da_dp = 1.0 / a * avec.dot(dav_dp)
        # print "d[a]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(da_dp, fd_grad[i]['da_dp'], i)
        assert approx_equal(da_dp, fd_grad[i]["da_dp"])

        db_dp = 1.0 / b * bvec.dot(dbv_dp)
        # print "d[b]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(db_dp, fd_grad[i]['db_dp'], i)
        assert approx_equal(db_dp, fd_grad[i]["db_dp"])

        dc_dp = 1.0 / c * cvec.dot(dcv_dp)
        # print "d[c]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(dc_dp, fd_grad[i]['dc_dp'], i)
        assert approx_equal(dc_dp, fd_grad[i]["dc_dp"])

        # print
        # print "CELL ANGLES"

        daa_dp = RAD2DEG * (dbv_dp.dot(dalpha_db) + dcv_dp.dot(dalpha_dc))
        dbb_dp = RAD2DEG * (dav_dp.dot(dbeta_da) + dcv_dp.dot(dbeta_dc))
        dcc_dp = RAD2DEG * (dav_dp.dot(dgamma_da) + dbv_dp.dot(dgamma_db))

        # print "d[alpha]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(daa_dp, fd_grad[i]['daa_dp'], i)
        # print "d[beta]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(dbb_dp, fd_grad[i]['dbb_dp'], i)
        # print "d[gamma]/dp{2} analytical: {0:.5f} FD: {1:.5f}".format(dcc_dp, fd_grad[i]['dcc_dp'], i)
        assert approx_equal(daa_dp, fd_grad[i]["daa_dp"])
        assert approx_equal(dbb_dp, fd_grad[i]["dbb_dp"])
        assert approx_equal(dcc_dp, fd_grad[i]["dcc_dp"])



"""
Restraints manager classes for scaling.
"""

from __future__ import annotations

from scitbx import sparse

from dials.array_family import flex


class ScalingRestraintsCalculator:
    """A class to calculate restraints for scaling for one or more datasets,
    by composition of the restraints of the individual datasets. The methods
    require a multi active_parameter_manager."""

    @staticmethod
    def calculate_restraints(multi_parameter_manager):
        """Calculate restraints for multi-dataset scaling, using a
        multi active_parameter_manager. Return None if no restraints, else return
        a restraints vector of length n_restrained_parameters and a gradient vector
        of length n_total_parameters."""
        R = flex.double([])
        G = flex.double([])
        for single_apm in multi_parameter_manager.apm_list:
            restr = SingleScalingRestraintsCalculator.calculate_restraints(single_apm)
            if restr:
                R.extend(restr[0])
                G.extend(restr[1])
            else:
                G.extend(flex.double(single_apm.n_active_params, 0.0))
        if R:
            return [R, G]
        return None

    @classmethod
    def calculate_jacobian_restraints(cls, multi_parameter_manager):
        """Calculate jacobian restraints for multi-dataset scaling, using a
        multi active_parameter_manager. First, the restraints are calculated - if
        not None is returned, then one restraints vector, jacobian matrix and
        weights vector is composed for the multiple datasets, else None is returned.
        The jacobian restraints matrix is of size n_restrained_parameters x
        n_parameters (across all datasets), while the residuals and weights vector
        are of length n_restrainted_parameters."""
        residual_restraints = cls.calculate_restraints(multi_parameter_manager)
        if residual_restraints:
            n_restraints = residual_restraints[0].size()
            weights = flex.double([])
            restraints_vector = flex.double([])
            jacobian = sparse.matrix(
                n_restraints, multi_parameter_manager.n_active_params
            )
            cumul_restr_pos = 0
            for i, single_apm in enumerate(multi_parameter_manager.apm_list):
                restraints = (
                    SingleScalingRestraintsCalculator.calculate_jacobian_restraints(
                        single_apm
                    )
                )
                if restraints:
                    jacobian.assign_block(
                        restraints[1],
                        cumul_restr_pos,
                        multi_parameter_manager.apm_data[i]["start_idx"],
                    )
                    cumul_restr_pos += restraints[1].n_rows
                    restraints_vector.extend(restraints[0])
                    weights.extend(restraints[2])
            return [restraints_vector, jacobian, weights]
        return None


class SingleScalingRestraintsCalculator:
    """A class to calculate restraints for scaling for an individual dataset, by
    using a single active_parameter_manager."""

    @classmethod
    def calculate_jacobian_restraints(cls, single_parameter_manager):
        """Calculate jacobian restraints for a single dataset from the scaling model
        components, using a single active_parameter_manager. Return None if no
        restraints, else return a residuals vector and a matrix of size
        n_restrained_parameters x n_parameters, and a weights vector."""
        residual_restraints = cls.calculate_restraints(single_parameter_manager)
        if residual_restraints:
            n_restraints = residual_restraints[0].size()
            weights = flex.double([])
            restraints_vector = flex.double([])
            jacobian = sparse.matrix(
                n_restraints, single_parameter_manager.n_active_params
            )
            cumul_restr_pos = 0
            for comp in single_parameter_manager.components.values():
                restraints = comp["object"].calculate_jacobian_restraints()
                if restraints:
                    jacobian.assign_block(
                        restraints[1], cumul_restr_pos, comp["start_idx"]
                    )
                    cumul_restr_pos += comp["n_params"]
                    restraints_vector.extend(restraints[0])
                    weights.extend(restraints[2])
            # Return the restraints vector, jacobian and weights (unity as weights
            # contained in individual jacobian/weoghts calculations).
            return [restraints_vector, jacobian, weights]
        return None

    @staticmethod
    def calculate_restraints(single_parameter_manager):
        """Calculate restraints for a single dataset from the scaling model
        components. Return None if no restraints, else return a residuals vector
        length n_restrained_parameters and a gradient vector of length n_parameters
        (of the scaling model for the individual dataset)."""
        residuals = flex.double([])
        gradient_vector = flex.double([])
        for comp in single_parameter_manager.components.values():
            resid = comp["object"].calculate_restraints()
            if resid:
                gradient_vector.extend(resid[1])
                residuals.extend(resid[0])
            else:
                gradient_vector.extend(flex.double(comp["n_params"], 0.0))
        if residuals:
            return [residuals, gradient_vector]
        return None



from __future__ import annotations

from unittest import mock

from boilerplate import run


def test_boilerplate(dials_data, tmpdir):
    x4wide = dials_data("x4wide_processed", pathlib=True)

    with tmpdir.as_cwd(), mock.patch("boilerplate.dials.util.log"):
        run(
            args=[
                str(x4wide / "AUTOMATIC_DEFAULT_scaled.expt"),
                str(x4wide / "AUTOMATIC_DEFAULT_scaled.refl"),
                "integer_parameter=42",
                "bool_parameter=True",
            ]
        )

    assert tmpdir.join("stronger.refl").check()



"""Auxiliary functions for the refinement package"""


from __future__ import annotations

import logging
import math
import random

import scitbx.matrix
from scitbx.array_family import flex

from dials_refinement_helpers_ext import CrystalOrientationCompose as xloc_cpp
from dials_refinement_helpers_ext import PanelGroupCompose as pgc_cpp
from dials_refinement_helpers_ext import dR_from_axis_and_angle as dR_cpp

logger = logging.getLogger(__name__)


def ordinal_number(array_index=None, cardinal_number=None):
    """Return a string representing the ordinal number for the input integer. One
    of array_index or cardinal_number must be set, depending on whether the
    input is from a 0-based or 1-based sequence.

    Based on Thad Guidry's post at
    https://groups.google.com/forum/#!topic/openrefine/G7_PSdUeno0"""
    if [array_index, cardinal_number].count(None) != 1:
        raise ValueError("One of array_index or cardinal_number should be set")
    if array_index is not None:
        i = int(array_index) + 1
    if cardinal_number is not None:
        i = int(cardinal_number)
    return str(i) + {1: "st", 2: "nd", 3: "rd"}.get(
        4 if 10 <= i % 100 < 20 else i % 10, "th"
    )


class PanelGroupCompose(pgc_cpp):
    """Wrapper for the C++ PanelGroupCompose class with accessors that
    return scitbx matrix values."""

    def d1(self):
        return scitbx.matrix.col(super().d1())

    def d2(self):
        return scitbx.matrix.col(super().d2())

    def origin(self):
        return scitbx.matrix.col(super().origin())

    def derivatives_for_panel(self, offset, dir1_new_basis, dir2_new_basis):
        d = super().derivatives_for_panel(offset, dir1_new_basis, dir2_new_basis)
        return [scitbx.matrix.sqr(e) for e in d]


class CrystalOrientationCompose(xloc_cpp):
    """Wrapper for the C++ CrystalOrientationCompose class with accessors that
    return matrix.sqr values."""

    def U(self):
        return scitbx.matrix.sqr(super().U())

    def dU_dphi1(self):
        return scitbx.matrix.sqr(super().dU_dphi1())

    def dU_dphi2(self):
        return scitbx.matrix.sqr(super().dU_dphi2())

    def dU_dphi3(self):
        return scitbx.matrix.sqr(super().dU_dphi3())


def dR_from_axis_and_angle(axis, angle, deg=False):
    """Wrapper for C++ version of dR_from_axis_and_angle returning a matrix.sqr"""
    return scitbx.matrix.sqr(dR_cpp(axis, angle, deg))


def dR_from_axis_and_angle_py(axis, angle, deg=False):
    """return the first derivative of a rotation matrix specified by its
    axis and angle"""

    # NB it is inefficient to do this separately from the calculation of
    # the rotation matrix itself, but it seems the Python interface to
    # scitbx does not have a suitable function. It might perhaps be
    # useful to write one, which could come straight from David Thomas'
    # RTMATS (present in Mosflm and MADNES).

    # NB RTMATS does calculation for a clockwise rotation of a vector
    # whereas axis_and_angle_as_r3_rotation_matrix does anticlockwise
    # rotation. Therefore flip the axis in here compared with
    # RTMATS in order to match the axis_and_angle_as_r3_rotation_matrix
    # convention

    # See also axis_and_angle_as_r3_derivative_wrt_angle, which does the same
    # as this function, but this function is faster.

    assert axis.n in ((3, 1), (1, 3))
    if deg:
        angle *= math.pi / 180
    axis = -1.0 * axis.normalize()
    ca, sa = math.cos(angle), math.sin(angle)

    return scitbx.matrix.sqr(
        (
            sa * axis[0] * axis[0] - sa,
            sa * axis[0] * axis[1] + ca * axis[2],
            sa * axis[0] * axis[2] - ca * axis[1],
            sa * axis[1] * axis[0] - ca * axis[2],
            sa * axis[1] * axis[1] - sa,
            sa * axis[1] * axis[2] + ca * axis[0],
            sa * axis[2] * axis[0] + ca * axis[1],
            sa * axis[2] * axis[1] - ca * axis[0],
            sa * axis[2] * axis[2] - sa,
        )
    )


def skew_symm(v):
    """Make matrix [v]_x from v. Essentially multiply vector by SO(3) basis
    set Lx, Ly, Lz. Equation (2) from Gallego & Yezzi paper.

    NB a C++ version exists in gallego_yezzi.h."""

    L1 = scitbx.matrix.sqr((0, 0, 0, 0, 0, -1, 0, 1, 0))
    L2 = scitbx.matrix.sqr((0, 0, 1, 0, 0, 0, -1, 0, 0))
    L3 = scitbx.matrix.sqr((0, -1, 0, 1, 0, 0, 0, 0, 0))

    v1, v2, v3 = v.elems

    return v1 * L1 + v2 * L2 + v3 * L3


def dRq_de(theta, e, q):
    """Calculate derivative of rotated vector r = R*q with respect to the elements
    of the rotation axis e, where the angle of rotation is theta.

    Implementation of Equation (8) from Gallego & Yezzi.

    NB a C++ version exists in gallego_yezzi.h."""

    # ensure e is unit vector
    e = e.normalize()

    # rotation matrix
    R = e.axis_and_angle_as_r3_rotation_matrix(theta, deg=False)

    # rotation vector v
    v = theta * e

    qx = skew_symm(q)
    vx = skew_symm(v)
    vvt = v * v.transpose()
    Rt = R.transpose()
    I3 = scitbx.matrix.identity(3)

    return (-1.0 / theta) * R * qx * (vvt + (Rt - I3) * vx)


def random_param_shift(vals, sigmas):
    """Add a random (normal) shift to a parameter set, for testing"""

    assert len(vals) == len(sigmas)
    shifts = [random.gauss(0, sd) for sd in sigmas]
    newvals = [(x + y) for x, y in zip(vals, shifts)]

    return newvals


def get_fd_gradients(mp, deltas, multi_state_elt=None):
    """Calculate centered finite difference gradients for each of the
    parameters of the model parameterisation mp.

    "deltas" must be a sequence of the same length as the parameter list, and
    contains the step size for the difference calculations for each parameter.

    "multi_state_elt" selects a particular state for use when mp is a multi-
    state parameterisation.
    """

    p_vals = mp.get_param_vals()
    assert len(deltas) == len(p_vals)
    fd_grad = []

    for i in range(len(deltas)):

        val = p_vals[i]

        p_vals[i] -= deltas[i] / 2.0
        mp.set_param_vals(p_vals)
        if multi_state_elt is None:
            rev_state = mp.get_state()
        else:
            rev_state = mp.get_state(multi_state_elt=multi_state_elt)

        p_vals[i] += deltas[i]
        mp.set_param_vals(p_vals)
        if multi_state_elt is None:
            fwd_state = mp.get_state()
        else:
            fwd_state = mp.get_state(multi_state_elt=multi_state_elt)

        fd_grad.append((fwd_state - rev_state) / deltas[i])

        p_vals[i] = val

    # return to the initial state
    mp.set_param_vals(p_vals)

    return fd_grad


def get_panel_groups_at_depth(group, depth=0):
    """Return a list of the panel groups at a certain depth below the node group"""
    assert depth >= 0
    if depth == 0:
        return [group]
    else:
        assert group.is_group()
        return [
            p
            for gp in group.children()
            for p in get_panel_groups_at_depth(gp, depth - 1)
        ]


def get_panel_ids_at_root(panel_list, group):
    """Get the sequential panel IDs for a set of panels belonging to a group"""
    if group.is_group():
        return [
            p for gp in group.children() for p in get_panel_ids_at_root(panel_list, gp)
        ]
    else:  # we got down to Panels
        return [panel_list.index(group)]


def string_sel(l, full_names, prefix=""):
    """Provide flexible matching between a list of input strings, l,
    consisting either of indices or partial names, and a list of full names,
    with an optional shared prefix. The input list l may arrive from PHIL
    conversion of the strings type. In that case, comma-separated values will
    require splitting, and bracket characters will be removed. The values in
    the processed list l should consist of integers or partial names. Integers
    will be treated as 0-based indices and partial names will be matched to
    as many full names as possible. The created selection is returned as a
    boolean list."""

    sel = [False] * len(full_names)
    full_names = [prefix + s for s in full_names]

    # expand elements of the list that are comma separated strings and remove
    # braces/brackets
    l = (s.strip("(){}[]") for e in l for s in str(e).split(","))
    l = (e for e in l if e != "")
    for e in l:
        try:
            i = int(e)
            sel[i] = True
            continue
        except ValueError:
            pass
        except IndexError:
            pass
        sel = [True if e in name else s for (name, s) in zip(full_names, sel)]

    return sel


def calculate_frame_numbers(reflections, experiments):
    """calculate observed frame numbers for all reflections, if not already
    set"""

    # Only do this if we have to
    if "xyzobs.px.value" in reflections:
        return reflections

    # Ok, frames are not set, so set them, with dummy observed pixel values
    frames = flex.double(len(reflections), 0.0)
    for iexp, exp in enumerate(experiments):
        scan = exp.scan
        if not scan:
            continue
        sel = reflections["id"] == iexp
        xyzobs = reflections["xyzobs.mm.value"].select(sel)
        angles = xyzobs.parts()[2]
        to_update = scan.get_array_index_from_angle(angles, deg=False)
        frames.set_selected(sel, to_update)
    reflections["xyzobs.px.value"] = flex.vec3_double(
        flex.double(len(reflections), 0.0), flex.double(len(reflections), 0.0), frames
    )

    return reflections


def set_obs_s1(reflections, experiments):
    """Set observed s1 vectors for reflections if required, return the number
    of reflections that have been set."""

    refs_wo_s1_sel = reflections["s1"].norms() < 1.0e-6
    nrefs_wo_s1 = refs_wo_s1_sel.count(True)
    if nrefs_wo_s1 == 0:
        return nrefs_wo_s1

    for i_expt, expt in enumerate(experiments):
        detector = expt.detector
        beam = expt.beam
        expt_sel = reflections["id"] == i_expt
        for i_panel, panel in enumerate(detector):
            panel_sel = reflections["panel"] == i_panel
            isel = (expt_sel & panel_sel & refs_wo_s1_sel).iselection()
            spots = reflections.select(isel)
            x, y, rot_angle = spots["xyzobs.mm.value"].parts()
            s1 = panel.get_lab_coord(flex.vec2_double(x, y))
            s1 = s1 / s1.norms() * (1 / beam.get_wavelength())
            reflections["s1"].set_selected(isel, s1)
    return nrefs_wo_s1



# LIBTBX_SET_DISPATCHER_NAME dials.complete_full_sphere


from __future__ import annotations

import logging
import math
import sys

from cctbx import crystal, miller
from dxtbx.model import ExperimentList
from libtbx.phil import parse
from scitbx import matrix

from dials.algorithms.refinement import rotation_decomposition
from dials.algorithms.shadowing.filter import filter_shadowed_reflections
from dials.array_family import flex
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, flatten_experiments

logger = logging.getLogger("dials.command_line.complete_full_sphere")

help_message = """
This program attempts to compute a sample realignment to measure the blind
region of reciprocal space, given an already recorded data set.

dials.complete_full_sphere [resolution=1.6] models.expt
"""

phil_scope = parse(
    """
resolution = 0.0
  .type = float
  .help = "Resolution of diffraction for blind region calculation"
shadow = True
  .type = bool
  .help = "Consider shadowing in calculating overall completeness"
"""
)


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        # The script usage
        usage = "usage: dials.complete_full_sphere [options] "

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            epilog=help_message,
            check_format=True,
            read_experiments=True,
        )

    def run(self, args=None):
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        log.config(logfile="dials.complete_full_sphere.log")

        model_shadow = params.shadow

        experiments = flatten_experiments(params.input.experiments)

        if len(experiments) != 1:
            self.parser.print_help()
            return

        expt = experiments[0]

        axes = expt.goniometer.get_axes()

        if len(axes) != 3:
            sys.exit("This will only work with 3-axis goniometers")

        if not expt.imageset.reader().get_format():
            sys.exit("This will only work with images available")

        if not expt.imageset.reader().get_format().get_goniometer_shadow_masker():
            model_shadow = False

        beam = expt.beam
        det = expt.detector

        if params.resolution:
            resolution = params.resolution
        else:
            resolution = det.get_max_inscribed_resolution(expt.beam.get_s0())

        # at this point, predict all of the reflections in the scan possible (i.e.
        # extend scan to 360 degrees) - this points back to expt

        self.make_scan_360(expt.scan)

        # now get a full set of all unique miller indices
        all_indices = miller.build_set(
            crystal_symmetry=crystal.symmetry(
                space_group=expt.crystal.get_space_group(),
                unit_cell=expt.crystal.get_unit_cell(),
            ),
            anomalous_flag=True,
            d_min=resolution,
        )

        if model_shadow:
            obs, shadow = self.predict_to_miller_set_with_shadow(expt, resolution)
        else:
            obs = self.predict_to_miller_set(expt, resolution)

        logger.info(
            "Fraction of unique observations at datum: %.1f%%",
            100.0 * len(obs.indices()) / len(all_indices.indices()),
        )

        missing = all_indices.lone_set(other=obs)

        logger.info("%d unique reflections in blind region", len(missing.indices()))

        e1 = matrix.col(axes[0])
        e2 = matrix.col(axes[1])
        e3 = matrix.col(axes[2])

        s0n = matrix.col(beam.get_s0()).normalize()

        # rotate blind region about beam by +/- two theta
        two_theta = 2.0 * math.asin(0.5 * beam.get_wavelength() / resolution)
        R_ptt = s0n.axis_and_angle_as_r3_rotation_matrix(two_theta)
        R_ntt = s0n.axis_and_angle_as_r3_rotation_matrix(-two_theta)

        # now decompose to e3, e2, e1
        sol_plus = rotation_decomposition.solve_r3_rotation_for_angles_given_axes(
            R_ptt, e3, e2, e1, return_both_solutions=True, deg=True
        )

        sol_minus = rotation_decomposition.solve_r3_rotation_for_angles_given_axes(
            R_ntt, e3, e2, e1, return_both_solutions=True, deg=True
        )

        solutions = []
        if sol_plus:
            solutions.extend(sol_plus)
        if sol_minus:
            solutions.extend(sol_minus)

        if not solutions:
            sys.exit(f"Impossible two theta: {two_theta * 180.0 / math.pi:.3f},")

        logger.info("Maximum two theta: %.3f,", two_theta * 180.0 / math.pi)
        logger.info("%d solutions found", len(solutions))

        names = tuple(
            [n.replace("GON_", "").lower() for n in expt.goniometer.get_names()]
        )

        logger.info(" %8s %8s %8s  coverage expt.expt" % names)
        self.write_expt(experiments, "solution_0.expt")
        for j, s in enumerate(solutions):
            expt.goniometer.set_angles(s)
            if model_shadow:
                obs, shadow = self.predict_to_miller_set_with_shadow(expt, resolution)
            else:
                obs = self.predict_to_miller_set(expt, resolution)
            new = missing.common_set(obs)
            fout = "solution_%d.expt" % (j + 1)
            f = len(new.indices()) / len(missing.indices())

            logger.info("%8.3f %8.3f %8.3f %4.2f %s", s[0], s[1], s[2], f, fout)
            self.write_expt(experiments, fout)

    def make_scan_360(self, scan):
        epochs = scan.get_epochs()
        exposure_times = scan.get_exposure_times()
        image_range = scan.get_image_range()
        oscillation = scan.get_oscillation()

        current = 1 + image_range[1] - image_range[0]
        turn = int(round(360.0 / oscillation[1]))
        extra = turn - current

        for j in range(extra):
            epochs.append(0.0)
            exposure_times.append(0.0)
        image_range = image_range[0], image_range[1] + extra

        scan.set_image_range(image_range)
        scan.set_epochs(epochs)
        scan.set_exposure_times(exposure_times)
        return scan

    def predict_to_miller_set(self, expt, resolution):
        predicted = flex.reflection_table.from_predictions(expt, dmin=resolution)
        hkl = predicted["miller_index"]

        # now get a full set of all unique miller indices
        obs = miller.set(
            crystal_symmetry=crystal.symmetry(
                space_group=expt.crystal.get_space_group(),
                unit_cell=expt.crystal.get_unit_cell(),
            ),
            anomalous_flag=True,
            indices=hkl,
        ).unique_under_symmetry()

        return obs

    def predict_to_miller_set_with_shadow(self, expt, resolution):
        predicted = flex.reflection_table.from_predictions(expt, dmin=resolution)

        # transmogrify this to an ExperimentList from an Experiment
        experiments = ExperimentList()
        experiments.append(expt)
        predicted["id"] = flex.int(predicted.size(), 0)
        shadowed = filter_shadowed_reflections(
            experiments, predicted, experiment_goniometer=True
        )
        predicted = predicted.select(~shadowed)

        hkl = predicted["miller_index"]

        # now get a full set of all unique miller indices
        obs = miller.set(
            crystal_symmetry=crystal.symmetry(
                space_group=expt.crystal.get_space_group(),
                unit_cell=expt.crystal.get_unit_cell(),
            ),
            anomalous_flag=True,
            indices=hkl,
        ).unique_under_symmetry()

        return obs, shadowed

    def write_expt(self, experiments, filename):
        experiments.as_file(filename)


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

from .fft1d import FFT1D
from .fft3d import FFT3D
from .real_space_grid_search import RealSpaceGridSearch
from .strategy import Strategy

__all__ = ["Strategy", "FFT1D", "FFT3D", "RealSpaceGridSearch"]



from __future__ import annotations

import contextlib
import inspect
import io
import os
import subprocess
import sys
from pathlib import Path

import libtbx
import libtbx.pkg_utils

try:
    import pkg_resources
except ModuleNotFoundError:
    pkg_resources = None

# So that we can import DIALS within this script, work out where the
# sources are and make them importable
_src_path_root = str(Path(libtbx.env.dist_path("dials")).joinpath("src"))
if _src_path_root not in sys.path:
    sys.path.insert(0, _src_path_root)

# Other packages, configured first, might have attempted to import dials, which
# would only get a namespace package, if the modules/ folder is configured as a
# repository. This means just setting the path wouldn't work. So, check to see
# if we have a namespace package imported, remove it (and any sub-packages),
# set the __path__, then import the real copy of DIALS.
#
# If this is *only* dxtbx, then we can probably get away without this by
# removing this part from dxtbx.
_dials = sys.modules.get("dials")
if _dials and _dials.__file__ is None:
    # Someone tried to import us and got a namespace package
    del sys.modules["dials"]
    # Remove any sub-modules that we might have tried and failed to import
    for _module in [x for x in sys.modules if x.startswith("dials.")]:
        del sys.modules[_module]

# Now, check to see if we configured XFEL first. If so, this is an error and we
# have a mis-configured environment.
if xfel_module := libtbx.env.module_dict.get("xfel"):
    dials_module = libtbx.env.module_dict.get("dials")
    if libtbx.env.module_list.index(xfel_module) < libtbx.env.module_list.index(
        dials_module
    ):
        sys.exit(
            """\033[31;1m
Error:
    The \033[34mxfel\033[31m module is loaded before the \033[34mdials\033[31m module in the libtbx
    module list. dials has changed to \033[0;1msrc/\033[1;31m layout, so in order for xfel
    to continue to find the headers it needs, it must be reconfigured.

    To fix this, please run:

        \033[0;1mlibtbx.configure xfel\033[0m

    \033[31;1mApologies for the inconvenience!\033[0m
"""
        )

import dials.precommitbx.nagger

try:
    from dials.util.version import dials_version

    print(dials_version())
except Exception:
    pass

dials.precommitbx.nagger.nag()

# During src/ transition, this ensures that the old entry points are cleared
libtbx.pkg_utils.define_entry_points({})


def _install_setup_readonly_fallback(package_name: str):
    """
    Partially install package in the libtbx build folder.
    This is a less complete installation - base python console_scripts
    entrypoints will not be installed, but the basic package metadata
    and other entrypoints will be enumerable through dispatcher black magic
    """
    root_path = libtbx.env.dist_path(package_name)
    import_path = os.path.join(root_path, "src")

    # Install this into a build/dxtbx subfolder
    build_path = abs(libtbx.env.build_path / package_name)
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--prefix",
            build_path,
            "--no-build-isolation",
            "--no-deps",
            "-e",
            root_path,
        ],
        check=True,
    )

    # Get the actual environment being configured (NOT libtbx.env)
    env = _get_real_env_hack_hack_hack()

    # Update the libtbx environment pythonpaths to point to the source
    # location which now has an .egg-info folder; this will mean that
    # the PYTHONPATH is written into the libtbx dispatchers
    rel_path = libtbx.env.as_relocatable_path(import_path)
    if rel_path not in env.pythonpath:
        env.pythonpath.insert(0, rel_path)

    # Update the sys.path so that we can find the .egg-info in this process
    # if we do a full reconstruction of the working set
    if import_path not in sys.path:
        sys.path.insert(0, import_path)

    # ...and add to the existing pkg_resources working_set
    if pkg_resources:
        pkg_resources.working_set.add_entry(import_path)

    # Add the src/ folder as an extra command_line_locations for dispatchers
    module = env.module_dict[package_name]
    if f"src/{package_name}" not in module.extra_command_line_locations:
        module.extra_command_line_locations.append(f"src/{package_name}")

    # Regenerate dispatchers for all modules with this new sys.path
    with contextlib.redirect_stdout(io.StringIO()):
        for module in env.module_list:
            module.process_command_line_directories()


def _get_real_env_hack_hack_hack():
    """
    Get the real, currently-being-configured libtbx.env environment.
    This is not libtbx.env, because although libtbx.env_config.environment.cold_start
    does:
        self.pickle()
        libtbx.env = self
    the first time there is an "import libtbx.load_env" this environment
    gets replaced by unpickling the freshly-written libtbx_env file onto
    libtbx.env, thereby making the environment accessed via libtbx.env
    *not* the actual one that is currently being constructed.
    So, the only way to get this environment being configured in order
    to - like - configure it, is to walk the stack trace and extract the
    self object from environment.refresh directly.
    """
    for frame in inspect.stack():
        if (
            frame.filename.endswith("env_config.py")
            and frame.function == "refresh"
            and "self" in frame.frame.f_locals
        ):
            return frame.frame.f_locals["self"]

    raise RuntimeError("Could not determine real libtbx.env_config.environment object")


def _create_dials_env_script():
    """
    write dials environment setup script and clobber cctbx setup scripts
    does nothing unless a file named 'dials'/'dials.bat' exists above
    the build/ directory
    """
    import os

    import libtbx.load_env

    if os.name == "nt":
        filename = abs(libtbx.env.build_path.dirname() / "dials.bat")
    else:
        filename = abs(libtbx.env.build_path.dirname() / "dials")

    if not os.path.exists(filename):
        return

    if os.name == "nt":
        script = """
rem enable conda environment
call %~dp0conda_base\\condabin\\activate.bat
rem prepend cctbx /build/bin directory to PATH
set PATH=%~dp0build\\bin;%PATH%
"""
    else:
        script = """
#!/bin/bash

if [ ! -z "${LIBTBX_BUILD_RELOCATION_HINT:-}" ]; then
  # possibly used for some logic in the installer
  LIBTBX_BUILD="${LIBTBX_BUILD_RELOCATION_HINT}"
  LIBTBX_BUILD_RELOCATION_HINT=
  export LIBTBX_BUILD_RELOCATION_HINT
elif [ -n "$BASH_SOURCE" ]; then
  LIBTBX_BUILD="$(dirname -- "${BASH_SOURCE[0]}")/build"
else
  LIBTBX_BUILD="%s"
fi

# make path absolute and resolve symlinks
LIBTBX_BUILD=$(cd -P -- "${LIBTBX_BUILD}" && pwd -P)

# enable conda environment
source ${LIBTBX_BUILD}/../conda_base/etc/profile.d/conda.sh
conda activate $(dirname -- "${LIBTBX_BUILD}")/conda_base

# prepend cctbx /build/bin directory to PATH
PATH="${LIBTBX_BUILD}/bin:${PATH}"
export PATH

# enable DIALS command line completion
[ -n "$BASH_VERSION" ] && {
  source $(libtbx.find_in_repositories dials/util/autocomplete.sh) && \
  source ${LIBTBX_BUILD}/dials/autocomplete/bash.sh || \
    echo dials command line completion not available
}

unset LIBTBX_BUILD
""" % abs(
            libtbx.env.build_path
        )
    with open(filename, "w") as fh:
        fh.write(script.lstrip())

    if os.name != "nt":
        mode = os.stat(filename).st_mode
        mode |= (mode & 0o444) >> 2  # copy R bits to X
        os.chmod(filename, mode)

    if os.name == "nt":
        clobber = """
echo {stars}
echo The script to set up the DIALS environment has changed
echo Please source or run {newscript} instead
echo {stars}
"""
        clobber_extensions = (".sh", ".csh", ".bat")
    else:
        clobber = """
echo '{stars}'
echo The script to set up the DIALS environment has changed
echo Please source or run '{newscript}' instead
echo '{stars}'
"""
        clobber_extensions = (".sh", ".csh", ".bat")

    for clobberfile_name in (
        "setpaths",
        "setpaths_all",
        "setpaths_debug",
    ):
        for clobber_ext in clobber_extensions:
            with open(
                abs(libtbx.env.build_path / (clobberfile_name + clobber_ext)), "w"
            ) as fh:
                fh.write(clobber.format(newscript=filename, stars="*" * 74))


def _install_dials_autocompletion():
    """generate bash.sh and SConscript file in /build/dials/autocomplete"""
    import os  # required due to cctbx weirdness

    import libtbx.load_env

    # Find the dials source directory
    dist_path = libtbx.env.dist_path("dials")

    # Set the location of the output directory
    output_directory = libtbx.env.under_build(os.path.join("dials", "autocomplete"))
    try:
        os.makedirs(output_directory)
    except OSError:
        pass

    # Build a list of autocompleteable commands
    commands_dir = os.path.join(dist_path, "src", "dials", "command_line")
    command_list = []
    for filename in sorted(os.listdir(commands_dir)):
        if not filename.startswith("_") and filename.endswith(".py"):
            # Check if this file marks itself as completable
            with open(os.path.join(commands_dir, filename), "rb") as f:
                if b"DIALS_ENABLE_COMMAND_LINE_COMPLETION" in f.read():
                    command_name = f"dials.{filename[:-3]}"
                    command_list.append(command_name)
    print("Identified autocompletable commands: " + " ".join(command_list))

    # Generate the autocompletion SConscript.
    with open(os.path.join(output_directory, "SConscript"), "w") as builder:
        builder.write(
            """import os.path
import libtbx.load_env\n
Import("env")\n\n
def dispatcher_outer(name):
    return os.path.join(libtbx.env.under_build("bin"), name)\n\n
def dispatcher_inner(name):
    return os.path.join(
        libtbx.env.dist_path("dials"), "src", "dials", "command_line", "%s.py" % name.partition(".")[2]
    )\n\n
env.Append(
    BUILDERS={{
        "AutoComplete": Builder(action="-$SOURCE --export-autocomplete-hints > $TARGET")
    }}
)
env["ENV"]["DIALS_NOBANNER"] = "1"
for cmd in [
{}
]:
    ac = env.AutoComplete(cmd, [dispatcher_outer(cmd), dispatcher_inner(cmd)])
    Requires(ac, Dir(libtbx.env.under_build("lib")))
    Depends(ac, os.path.join(libtbx.env.dist_path("dials"), "src", "dials", "util", "options.py"))
    Depends(ac, os.path.join(libtbx.env.dist_path("dials"), "util", "autocomplete.sh"))
""".format(
                "\n".join([f'    "{cmd}",' for cmd in command_list])
            )
        )

    # Generate a bash script activating command line completion for each relevant command
    with open(os.path.join(output_directory, "bash.sh"), "w") as script:
        script.write("type compopt &>/dev/null && {\n")
        for cmd in command_list:
            script.write(f" complete -F _dials_autocomplete {cmd}\n")
        script.write("}\n")
        script.write("type compopt &>/dev/null || {\n")
        for cmd in command_list:
            script.write(f" complete -o nospace -F _dials_autocomplete {cmd}\n")
        script.write("}\n")


# _install_package_setup("dials")
_install_setup_readonly_fallback("dials")
_create_dials_env_script()
_install_dials_autocompletion()
